/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <ctype.h>
#include <math.h>

#define BIG_MAX 0x3FFFFFFF
#define BIG_RADIX 0x40000000
#define LOG_BIG_RADIX 30
#define DEC_PER_BIG_DIGIT 9
#define SMALL_NUM_STR_LEN 8 /* conservative is OK */

Scheme_Object *scheme_make_small_bignum(int v, Small_Bignum *o)
{
  if (!o)
    o = (Small_Bignum *)scheme_malloc(sizeof(Small_Bignum));

  o->o.type = scheme_bignum_type;
  if (!v)
    SCHEME_PINT_VAL(&o->o) = 0;
  else
    SCHEME_PINT_VAL(&o->o) = v < 0 ? -1 : 1;
  SCHEME_IPTR_VAL(&o->o) = (void *)o->v;

  if (v < 0)
    v = -v;

  o->v[0] = v;

  return (Scheme_Object *)o;
}

Scheme_Object *scheme_make_bignum(int v)
{
  return scheme_make_small_bignum(v, NULL);
}

Scheme_Object *scheme_bignum_normalize(Scheme_Object *o)
{
  int v;
  
  if (!SCHEME_BIGNUMP(o))
    return o;

  v = SCHEME_PINT_VAL(o);
  if (!v)
    return scheme_make_integer(0);
  if (v == 1)
    return scheme_make_integer(*(bignum *)SCHEME_IPTR_VAL(o));
  if (v == -1)
    return scheme_make_integer(-(*(bignum *)SCHEME_IPTR_VAL(o)));

  return o;
}

int scheme_bignum_eq(Scheme_Object *a, Scheme_Object *b)
{
  int al, bl;
  bignum *aa, *ba;

  al = SCHEME_PINT_VAL(a);
  bl = SCHEME_PINT_VAL(b);
  if (al != bl)
    return 0;

  aa = (bignum *)SCHEME_IPTR_VAL(a);
  ba = (bignum *)SCHEME_IPTR_VAL(b);
  
  while (al--)
    if (*(aa++) != *(ba++))
      return 0;

  return 1;
}

int scheme_bignum_lt(Scheme_Object *a, Scheme_Object *b)
{
  int al, bl;
  bignum *aa, *ba, av, bv;

  al = SCHEME_PINT_VAL(a);
  bl = SCHEME_PINT_VAL(b);
  if (al != bl)
    return (al < bl);

  if (al < 0) {
    bignum *save;

    al = -al;

    save = aa;
    aa = ba;
    ba = save;
  }

  aa = ((bignum *)SCHEME_IPTR_VAL(a)) + al;
  ba = ((bignum *)SCHEME_IPTR_VAL(b)) + al;
  
  while (al--) {
    av = *(--aa);
    bv = *(--ba);
    if (av != bv)
      return (av < bv);
  }

  return 0;
}

int scheme_bignum_gt(Scheme_Object *a, Scheme_Object *b)
{
  return scheme_bignum_lt(b, a);
}

int scheme_bignum_le(Scheme_Object *a, Scheme_Object *b)
{
  return !scheme_bignum_gt(a, b);
}

int scheme_bignum_ge(Scheme_Object *a, Scheme_Object *b)
{
  return !scheme_bignum_lt(a, b);
}

Scheme_Object *scheme_bignum_negate(Scheme_Object *n)
{
  Scheme_Object *o;

  o = (Scheme_Object *)scheme_malloc(sizeof(Scheme_Object));

  o->type = scheme_bignum_type;
  SCHEME_PINT_VAL(o) = -SCHEME_PINT_VAL(n);
  SCHEME_IPTR_VAL(o) = SCHEME_IPTR_VAL(n);
 
  return o;
}

void bignum_negate_inplace(Scheme_Object *n)
{
  SCHEME_PINT_VAL(n) = -SCHEME_PINT_VAL(n);
}

Scheme_Object *bignum_copy(Scheme_Object *n, int copy_array)
{
  Scheme_Object *o;
  int c, cnt;

  c = cnt = SCHEME_PINT_VAL(n);

  if (cnt < 0)
    cnt = -cnt;

  o = (Scheme_Object *)scheme_malloc(sizeof(Scheme_Object));

  o->type = scheme_bignum_type;
  SCHEME_PINT_VAL(o) = c;
  if (copy_array) {
    SCHEME_IPTR_VAL(o) = scheme_malloc_atomic(sizeof(bignum) * cnt);
    memcpy(SCHEME_IPTR_VAL(o), SCHEME_IPTR_VAL(n), sizeof(bignum) * cnt);
  } else {
    SCHEME_IPTR_VAL(o) = SCHEME_IPTR_VAL(n);
  }

  return o;
}

static Scheme_Object *bignum_add(Scheme_Object *o, bignum **buffer, int *size,
				 Scheme_Object *a, Scheme_Object *b, int norm)
{
  int al, bl, nal, nbl, vl, nvl, i, same;
  long s, av, bv;
  bignum *aa, *ba, *va, carry;

  al = SCHEME_PINT_VAL(a);
  bl = SCHEME_PINT_VAL(b);

  nal = (al < 0) ? -al : al;
  nbl = (bl < 0) ? -bl : bl;

  aa = ((bignum *)SCHEME_IPTR_VAL(a));
  ba = ((bignum *)SCHEME_IPTR_VAL(b));
  
  same = ((al > 0) == (bl > 0));

  nvl = ((nal > nbl) ? nal : nbl);
 
  if (!same) {
    int reverse, sl;
    bignum *sa;

    if (nal == nbl) {
      reverse = 0;
      for (i = nal; i--; )
	if (aa[i] != ba[i]) {
	  reverse = (aa[i] < ba[i]);
	  break;
	}
    } else
      reverse = (nal < nbl);
    
    if (reverse) {
      sl = nal;
      nal = nbl;
      nbl = sl;
      sl = al;
      al = bl;
      bl = sl;
      sa = aa;
      aa = ba;
      ba = sa;
    }
  } else
    nvl++;
    
  if (buffer) {
    if (*size < nvl) {
      *size = 2 * nvl;
      *buffer = (bignum *)scheme_malloc_atomic(*size * sizeof(bignum));
    }
    va = *buffer;
  } else
    va = (bignum *)scheme_malloc_atomic(nvl * sizeof(bignum));
  
  carry = 0;
  for (i = 0; i < nvl; i++) {
    av = (i >= nal) ? 0 : aa[i];
    bv = (i >= nbl) ? 0 : ba[i];
      
    if (same)
      s = (long)av + (long)bv + carry;
    else
      s = (long)av - (long)bv + carry;
    
    if (s < 0) {
      carry = -1;
      s += BIG_RADIX;
    } else {
      carry = (s & BIG_RADIX) ? 1 : 0;
      s = s & BIG_MAX;
    }
    
    va[i] = s;
  }

  /* Get rid of leading zeros: */
  for (i = nvl; i--; )
    if (va[i])
      break;
  vl = i + 1;

  if (al < 0)
    vl = -vl;
  
  if (!o) {
    o = scheme_alloc_object();
    o->type = scheme_bignum_type;
  }
  SCHEME_PINT_VAL(o) = vl;
  SCHEME_IPTR_VAL(o) = (void *)va;

  if (norm)
    return scheme_bignum_normalize(o);
  else
    return o;
}

Scheme_Object *scheme_bignum_add(Scheme_Object *a, Scheme_Object *b)
{
  return bignum_add(NULL, NULL, NULL, a, b, 1);
}

Scheme_Object *scheme_bignum_subtract(Scheme_Object *a, Scheme_Object *b)
{
  return scheme_bignum_add(a, scheme_bignum_negate(b));
}

Scheme_Object *scheme_bignum_add1(Scheme_Object *n)
{
  static Scheme_Object *one = NULL;

  if (!one)
    one = scheme_make_bignum(1);

  return scheme_bignum_add(n, one);
}

Scheme_Object *scheme_bignum_sub1(Scheme_Object *n)
{
  static Scheme_Object *none = NULL;

  if (!none)
    none = scheme_make_bignum(-1);

  return scheme_bignum_add(n, none);
}

static void bignum_double_inplace(Scheme_Object *n)
{
  int nl, i;
  bignum *na, *naya, carry;

  /* Because bignum calculations are not bounded: */
  if ((scheme_fuel_counter -= 5) < 0) {
    scheme_process_block(0);
    scheme_fuel_counter = scheme_current_process->engine_weight;
  }

  nl = SCHEME_PINT_VAL(n);
  if (nl < 0)
    nl = -nl;
  na = (bignum *)SCHEME_IPTR_VAL(n);
  
  carry = 0;
  for (i = 0; i < nl; i++) {
    na[i] = (na[i] << 1) + carry;
    carry = (na[i] & BIG_RADIX) ? 1 : 0;
    if (carry)
      na[i] = na[i] & BIG_MAX;
  }

  if (carry) {
    /* expand */
    naya = (bignum *)scheme_malloc_atomic(sizeof(bignum) * (nl + 1));
    for (i = 0; i < nl; i++)
      naya[i] = na[i];
    naya[nl] = 1;

    SCHEME_IPTR_VAL(n) = (void *)naya;

    if (SCHEME_PINT_VAL(n) < 0)
      SCHEME_PINT_VAL(n) -= 1;
    else
      SCHEME_PINT_VAL(n) += 1;
  }
}

static void bignum_half_inplace(Scheme_Object *n)
{
  int nl, i;
  bignum *na, carry, next;
  
  nl = SCHEME_PINT_VAL(n);
  if (nl < 0)
    nl = -nl;
  na = (bignum *)SCHEME_IPTR_VAL(n);
  
  carry = 0;
  for (i = nl; i--; ) {
    next = (na[i] & 1) ? (BIG_RADIX >> 1) : 0;
    na[i] = (na[i] >> 1) + carry;
    carry = next;
  }

  if (nl && !na[nl - 1]) {
    --nl;
    if (SCHEME_PINT_VAL(n) < 0)
      SCHEME_PINT_VAL(n) = -nl;
    else
      SCHEME_PINT_VAL(n) = nl;
  }
}

Scheme_Object *bignum_multiply(Small_Bignum *rsmall,
			       Scheme_Object *a, Scheme_Object *b, int norm)
{
  Scheme_Object *r;
  int bl, i, negate;
  int size = 0;
  bignum *ba, v, *buffer;

  if (scheme_bignum_lt(a, b)) {
    r = a;
    a = b;
    b = r;
  }

  bl = SCHEME_PINT_VAL(b);
  if (bl < 0) {
    bl = -bl;
    negate = 1;
  } else
    negate = 0;
  ba = (bignum *)SCHEME_IPTR_VAL(b);
  ba += bl;

  r = scheme_make_small_bignum(0, rsmall);

  while (bl--) {
    v = *(--ba);

    for (i = 0; i < LOG_BIG_RADIX; i++) {
      bignum_double_inplace(r);
      v = v << 1;
      if (v & BIG_RADIX)
	r = bignum_add(r, &buffer, &size, r, a, 0);
    }
  }

  if (negate)
    bignum_negate_inplace(r);

  if (norm) {
    r = scheme_bignum_normalize(r);
    if (SCHEME_BIGNUMP(r))
      r = bignum_copy(r, 0);
  }

  return r;
}

Scheme_Object *scheme_bignum_multiply(Scheme_Object *a, Scheme_Object *b)
{
  return bignum_multiply(NULL, a, b, 1);
}

Scheme_Object *scheme_bignum_power(Scheme_Object *a, Scheme_Object *b)
{
  Small_Bignum s1, s2;
  Scheme_Object *r;
  int bl, i, toggle = 0;
  bignum *ba, v;

  bl = SCHEME_PINT_VAL(b);
  if (bl < 0) {
    scheme_signal_error("positive bignum powers only");
  }
  ba = (bignum *)SCHEME_IPTR_VAL(b);
  ba += bl;

  r = scheme_make_small_bignum(1, &s1);
  while (bl--) {
    v = *(--ba);

    for (i = 0; i < LOG_BIG_RADIX; i++) {
      r = bignum_multiply(toggle ? &s1 : &s2, r, r, 0);
      toggle = !toggle;
      v = v << 1;
      if (v & BIG_RADIX) {
	r = bignum_multiply(toggle ? &s1 : &s2, r, a, 0);
	toggle = !toggle;
      }
    }
  }

  r = scheme_bignum_normalize(r);

  if (SCHEME_BIGNUMP(r))
    r = bignum_copy(r, 0);

  return r;
}

Scheme_Object *scheme_bignum_max(Scheme_Object *a, Scheme_Object *b)
{
  return scheme_bignum_normalize((scheme_bignum_lt(a, b)) ? b : a);
}

Scheme_Object *scheme_bignum_min(Scheme_Object *a, Scheme_Object *b)
{
  return scheme_bignum_normalize((scheme_bignum_lt(a, b)) ? a : b);
}

double scheme_bignum_to_float(Scheme_Object *n)
{
  int nl;
  bignum *na;
  double d;

  nl = SCHEME_PINT_VAL(n);
  if (nl < 0)
    nl = -nl;
  na = (bignum *)SCHEME_IPTR_VAL(n) + nl;

  d = 0;
  while (nl--) {
    d *= (float)BIG_RADIX;
    d += *(--na);
  }
  
  if (SCHEME_PINT_VAL(n) < 0)
    d = -d;

  return d;
}

char *scheme_bignum_to_string(Scheme_Object *b, int radix)
{
  char *n, *naya;
  int size, i, j, old, top;
  int bl, carry;
  bignum *ba, v;

  if (radix != 10 && radix != 2 && radix != 8 && radix != 16)
    scheme_signal_error("bad bignum radix (%d)", radix);

  size = 10;
  n = scheme_malloc_atomic(size);

  for (i = 0; i < size; i++)
    n[i] = '0';

  bl = SCHEME_PINT_VAL(b);
  if (bl < 0)
    bl = -bl;
  ba = (bignum *)SCHEME_IPTR_VAL(b);
  ba += bl;

  top = radix + '0' - 1;

  while (bl--) {
    v = *(--ba);

    for (j = 0; j < LOG_BIG_RADIX; j++) {
      /* Double it */
      carry = 0;
      for (i = 0; i < size; i++) {
	n[i] += carry + (n[i] - '0');
	if (n[i] > top) {
	  carry = 1;
	  n[i] -= radix;
	} else
	  carry = 0;
      }
      if (carry) {
	/* Expand: */
	old = size;
	size *= 2;
	naya = scheme_malloc_atomic(size);
	  
	for (i = 0; i < old; i++)
	  naya[i] = n[i];
	naya[i++] = carry + '0';
	for (; i < size; i++)
	  naya[i] = '0';

	n = naya;
      }

      v = v << 1;

      /* Add one */
      if (v & BIG_RADIX) {
	for (i = 0; i < size; i++) {
	  if (n[i] < top) {
	    n[i]++;
	    break;
	  } else
	    n[i] = '0';
	}
	if (i >= size) {
	  /* Expand: */
	  old = size;
	  size *= 2;
	  n = scheme_malloc_atomic(size);
	  
	  for (i = 0; i < size; i++)
	    n[i] = '0';
	  n[old] = '1';
	}
      }
    }
  }

  for (i = size; i--; )
    if (n[i] != '0')
      break;
  size = i + 1;
  if (!size)
    size = 1;

  if (radix > 10) {
    for (i = size; i--; )
      if (n[i] > '9')
	n[i] = (n[i] - '9') + 'A' - 1;
  }

  i = size;
  if (SCHEME_PINT_VAL(b) < 0)
    i++;

  naya = scheme_malloc_atomic(i + 1);
  naya[i] = 0;

  if (i != size) {
    naya[0] = '-';
    i = 1;
  } else
    i = 0;
    
  n += size;
  for (j = 0; j < size; j++, i++)
    naya[i] = *(--n);

  return naya;
}

Scheme_Object *decimal_digits[17];

Scheme_Object *scheme_read_bignum(const char *str, int radix)
{
  Small_Bignum s1, s2;
  Scheme_Object *r;
  int size = 0, read_valid;
  int len, i, d, negate;
  bignum *buffer;

  if (!decimal_digits[0])
    for (i = 0; i < 17; i++)
      decimal_digits[i] = scheme_make_bignum(i);

  negate = 0;
  while (isspace(*str) || (*str == '+') || (*str == '-')) {
    if (*str == '-')
      negate = !negate;
    str++;
  }

  len = strlen(str);

  if (radix == 10 && (len < SMALL_NUM_STR_LEN)) {
    if (!*str)
      return scheme_false;
    for (i = 0; *str; str++) {
      if (*str < '0' || *str > '9')
	return scheme_false;
      i = (i * 10) + (*str - '0');
    }
    if (negate)
       i = -i;
    return scheme_make_integer(i);
  }

  r = scheme_make_small_bignum(0, &s2);

  read_valid = 0;

  while (len--) {
    d = *(str++);
    if (isspace(d))
      break;
    if (d >= '0' && d <= '9')
      d -= '0';
    else if (d >= 'a' && d <= 'z')
      d -= 'a' - 10;
    else if (d >= 'A' && d <= 'Z')
      d -= 'A' - 10;
    else
      return scheme_false;
    if (d < 0 || d >= radix)
      return scheme_false;
    read_valid = 1;
    r = bignum_multiply(&s1, r, decimal_digits[radix], 0);
    r = bignum_add((Scheme_Object *)&s2, &buffer, &size, 
		   r, decimal_digits[d], 0);
  }

  if (!read_valid)
    /* No digits */
    return scheme_false;

  if (negate)
    bignum_negate_inplace(r);

  r = scheme_bignum_normalize(r);
  if (SCHEME_BIGNUMP(r))
    r = bignum_copy(r, 0);

  return r;
}

#define USE_FLOAT_BITS 50

Scheme_Object *scheme_bignum_from_float(double d)
{
  Small_Bignum s1, s2, onesmall;
  int negate, log, times, i, size = 0;
  double r;
  Scheme_Object *n, *m, *one;
  int toggle = 0;
  bignum *buffer;

  if (d < 0) {
    negate = 1;
    d = -d;
  } else
    negate = 0;

  if (d < 1.0)
    return scheme_make_integer(0);

  r = 1;
  log = 0;
  while (r < d) {
    log++;
    r *= 2.0;
  }

  if (log > USE_FLOAT_BITS) {
    times = log - USE_FLOAT_BITS;
    log = USE_FLOAT_BITS;
    for (i = 0; i < times; i++)
      d /= 2;
  } else
    times = 0;

  r = pow(2.0, log);

  n = scheme_make_small_bignum(0, &s1);
  one = scheme_make_small_bignum(1, &onesmall);

  log++;
  while (log--) {
    bignum_double_inplace(n);
    if (d >= r) {
      d -= r;
      n = bignum_add(n, &buffer, &size, n, one, 0);
    }
    r /= 2;
  }

  if (times) {
    m = scheme_make_bignum(1);
    while (times--)
      bignum_double_inplace(m);      
    n = bignum_multiply(toggle ? &s1 : &s2, n, m, 0);
    toggle = !toggle;
  }

  if (negate)
    bignum_negate_inplace(n);

  n = scheme_bignum_normalize(n);
  if (SCHEME_BIGNUMP(n))
    n = bignum_copy(n, 0);

  return n;
}

void scheme_bignum_divide(Scheme_Object *n, Scheme_Object *d,
			  Scheme_Object **qp, Scheme_Object **rp, int norm)
{
  Scheme_Object *m, *q, *r, *one;
  int log, negate_q, rsize = 0, qsize = 0;
  bignum *rbuffer, *qbuffer;

  negate_q = 0;
  if (SCHEME_PINT_VAL(n) < 0) {
    n = scheme_bignum_negate(n);
    negate_q = !negate_q;
  }
  if (SCHEME_PINT_VAL(d) < 0) {
    d = scheme_bignum_negate(d);
    negate_q = !negate_q;
  }

  if (scheme_bignum_gt(d, n)) {
    if (qp)
      *qp = norm ? scheme_make_integer(0) : scheme_make_bignum(0);
    if (rp)
      *rp = norm ? scheme_bignum_normalize(n) : n;
    return;
  }

  log = 0;
  m = bignum_copy(d, 1);
  while (scheme_bignum_lt(m, n)) {
    bignum_double_inplace(m);
    log++;
  }

  one = scheme_make_bignum(1);

  q = scheme_make_bignum(0);
  r = bignum_copy(n, 1);
  log++;
  while (log--) {
    bignum_double_inplace(q);
    if (!scheme_bignum_lt(r, m)) {
      bignum_negate_inplace(m);
      r = bignum_add(r, &rbuffer, &rsize, r, m, 0);
      bignum_negate_inplace(m);
      q = bignum_add(q, &qbuffer, &qsize, q, one, 0);
    }
    bignum_half_inplace(m);
  }

  if (qp) {
    if (negate_q)
      bignum_negate_inplace(q);
    *qp = norm ? scheme_bignum_normalize(q) : q;
  }
  if (rp)
    *rp = norm ? scheme_bignum_normalize(r) : r;
}

