/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"

/* globals */
void (*scheme_console_printf)(char *str, ...);
void (*scheme_exit)(int v);
int scheme_error_takes_symbol;

/* locals */
static Scheme_Object *error(int argc, Scheme_Object *argv[]);
static Scheme_Object *error_handler(int argc, Scheme_Object *argv[]);
static Scheme_Object *exit_handler(int argc, Scheme_Object *argv[]);
static Scheme_Object *def_error_handler_proc(int argc, Scheme_Object *argv[]);
static Scheme_Object *def_exit_handler_proc(int argc, Scheme_Object *argv[]);
static Scheme_Object *sch_exit (int argc, Scheme_Object *argv[]);

static Scheme_Object *newline_char, *default_exit_handler;

static void default_printf(char *msg, ...)
{
  va_list args;
  va_start(args, msg);
  vfprintf(stderr, msg, args);
  va_end(args);
}

void scheme_init_process_error(Scheme_Process *p)
{
  p->error_handler_proc = 
    scheme_make_prim_w_arity(def_error_handler_proc, 2, -1);
}

void
scheme_init_error (Scheme_Env *env)
{
  if (!scheme_console_printf)
    scheme_console_printf = default_printf;

  scheme_add_global_constant("error", 
			     scheme_make_prim_w_arity(error, 1, -1), 
			     env);
  scheme_add_global_constant("error-handler", 
			     scheme_make_prim_w_arity(error_handler, 0, 1), 
			     env);
  scheme_add_global_constant("exit-handler", 
			     scheme_make_prim_w_arity(exit_handler, 1, 1), 
			     env);
  scheme_add_global_constant("exit", 
			     scheme_make_prim_w_arity(sch_exit, 0, 1), 
			     env);

  if (scheme_starting_up) {
    newline_char = scheme_make_char('\n');
    default_exit_handler = 
      scheme_make_prim_w_arity(def_exit_handler_proc, 1, 1);
  }

  scheme_init_process_error(scheme_current_process);
}

static char *buffer;
  
void
scheme_signal_error (char *msg, ...)
{
  va_list args;
  Scheme_Object *p[2];
  jmp_buf savebuf;
  static displaying = 0;

  if (!buffer)
    buffer = scheme_malloc_atomic(1000);

  va_start (args, msg);
  vsprintf (buffer, msg, args);
  va_end (args);

  if (scheme_in_macro_application) {
    scheme_in_macro_application = 0;
    strcat(buffer, " [macro apply]");
  }

  if (scheme_starting_up) {
    scheme_console_printf("%s", buffer);
    exit(0);
  }

  if (displaying) {
    scheme_console_printf("Error trying to display error: %s\n", buffer);
    scheme_longjmp (scheme_error_buf, 1);
  } else {
    displaying = 1;
    p[0] = scheme_debug_get_last_eval();
    p[1] = scheme_make_string(buffer);
    memcpy(&savebuf, &scheme_error_buf, sizeof(jmp_buf));
    if (setjmp(scheme_error_buf)) {
      displaying = 0;
      scheme_longjmp(savebuf, 1);
    } else {
      scheme_apply(scheme_current_process->error_handler_proc, 2, p);
      /* Typically jumps out of here */
      displaying = 0;
      scheme_longjmp(savebuf, 1); /* force an exit */
    }
  }
}

void
scheme_warning (char *msg, ...)
{
  va_list args;

  if (!buffer)
    buffer = scheme_malloc_atomic(1000);

  va_start (args, msg);
  vsprintf (buffer, msg, args);
  va_end (args);

  strcat(buffer, "\n");
  scheme_display(scheme_make_string(buffer), scheme_stderr_port);
}

void scheme_wrong_count(const char *name, int minc, int maxc, int argc,
			Scheme_Object **argv)
{
  if (!maxc)
    scheme_signal_error("%s: expects no arguments, given %d",
			name, argc);
  else if (maxc < 0)
    scheme_signal_error("%s: expects at least %d arguments, given %d",
			name, minc, argc);
  else if (minc == maxc)
    scheme_signal_error("%s: expects %d arguments, given %d",
			name, minc, argc);
  else
    scheme_signal_error("%s: expects %d to %d arguments, given %d",
			name, minc, maxc, argc);
}

void scheme_wrong_type(const char *name, const char *expected, 
		       int which, int argc,
		       Scheme_Object **argv)
{
#define MAX_VAL_LEN 20

  Scheme_Object *o;
  char *s, *type;
  long size;
  static char *ending[] = {"st", "nd", "rd"};

  o = argv[which < 0 ? 0 : which];
  type = scheme_get_type_name(SCHEME_TYPE(o));

  s = scheme_write_to_string(o, &size);
  if (size > MAX_VAL_LEN) {
    s[MAX_VAL_LEN - 3] = '.';
    s[MAX_VAL_LEN - 2] = '.';
    s[MAX_VAL_LEN - 1] = '.';
    s[MAX_VAL_LEN] = 0;
  }

  if (which < 0)
    scheme_signal_error("%s: expects args of type <%s>; "
			"given %s (type %s)",
			name, expected, s, type);
  else
    scheme_signal_error("%s: expects type <%s> as %d%s arg; "
			"given %s (type %s)",
			name, expected, which + 1,
			((which < 10 || which > 20)
			 && ((which % 10) < 3)) ? ending[which % 10] : "th",
			s, type);
}

static Scheme_Object *error(int argc, Scheme_Object *argv[])
{
  Scheme_Object **newargs;

  if (!argc)
    scheme_wrong_count("error", 1, -1, argc, argv);
  if (SCHEME_SYMBOLP(argv[0])) {
    if (argc < 2) {
      /* Just a symbol */
      newargs = (Scheme_Object**)scheme_malloc(2*sizeof(Scheme_Object*));
      newargs[1] = 
	scheme_append_string(scheme_make_string("error: "),
			     scheme_make_string(SCHEME_SYM_VAL(argv[0])));
    } else {
      Scheme_Object *format, *sym, *msg;

      /* Chez-style: symbol, format string, format items... */
      if (!SCHEME_STRINGP(argv[1]))
	scheme_signal_error("error", "string", 1, argc, argv);
      
      format = scheme_lookup_global(scheme_intern_symbol("format"),
				    scheme_env);

      if (!format)
	scheme_signal_error("error: can't find format procedure"
			    " to handle error");

      sym = argv[0];
      argv[0] = scheme_append_string(scheme_make_string("~s: "),
				     argv[1]);
      argv[1] = sym;

      msg = scheme_apply(format, argc, argv);
      if (!SCHEME_STRINGP(msg))
	scheme_signal_error("error: format procedure returned a non-string");

      newargs = (Scheme_Object**)scheme_malloc(2*sizeof(Scheme_Object*));
      newargs[1] = msg;
    }
    argc = 2;
  } else {
    /* String followed by other values: */
    if (!SCHEME_STRINGP(argv[0]))
      scheme_wrong_type("error", "string", 0, argc, argv);
    
    newargs = (Scheme_Object**)scheme_malloc((argc+1)*sizeof(Scheme_Object*));
    memcpy(newargs + 1, argv, argc * sizeof(Scheme_Object*));

    argc++;
  }

  newargs[0] = scheme_debug_get_last_eval();

  return _scheme_tail_apply(scheme_current_process->error_handler_proc, 
			    argc, newargs);
}

static Scheme_Object *
def_error_handler_proc(int argc, Scheme_Object *argv[])
{
  int i;

  if (argc < 2)
    scheme_wrong_count("default-error-handler-proc", 2, -1, argc, argv);
  if (!SCHEME_DEBUGP(argv[0]))
    scheme_wrong_type("default-error-handler-proc", "debug handle", 0,
		      argc, argv);
  if (!SCHEME_STRINGP(argv[1]))
    scheme_wrong_type("default-error-handler-proc", "string", 1, argc, argv);

  scheme_display (argv[1], scheme_stderr_port);
  for (i = 2; i < argc ; i++) {
    scheme_write_string(" ", 1, scheme_stderr_port);
    scheme_write (argv[i], scheme_stderr_port);
  }
  scheme_display(newline_char, scheme_stderr_port);
  scheme_longjmp(scheme_error_buf, 1);
}

static Scheme_Object *
error_handler(int argc, Scheme_Object *argv[])
{
  if (argc >= 2)
    scheme_wrong_count("error-handler", 0, 1, argc, argv);
  
  if (argc) {
    if (!SCHEME_PROCP(argv[0]))
      scheme_wrong_type("error-handler", "closure", 0, argc, argv);
    scheme_current_process->error_handler_proc = argv[0];
    return scheme_void;
  } else
    return scheme_current_process->error_handler_proc;
}

static Scheme_Object *
exit_handler(int argc, Scheme_Object *argv[])
{
  if (argc >= 2)
    scheme_wrong_count("exit-handler", 0, 1, argc, argv);
  
  if (argc) {
    if (!SCHEME_PROCP(argv[0]))
      scheme_wrong_type("exit-handler", "closure", 0, argc, argv);
    scheme_current_process->exit_handler = argv[0];
    return scheme_void;
  } else if (scheme_current_process->exit_handler)
    return scheme_current_process->exit_handler;
  else
    return default_exit_handler;
}

static Scheme_Object *
def_exit_handler_proc(int argc, Scheme_Object *argv[])
{
  int status;

  if (argc != 1)
    scheme_wrong_count("default-exit-handler", 1, 1, argc, argv);

  if (!SCHEME_INTP (argv[0]))
    scheme_wrong_type("default-exit-handler", "integer", 0, argc, argv);

  status = SCHEME_INT_VAL(argv[0]);

  if (scheme_exit)
    scheme_exit(status);
  else
    exit(status);

  return scheme_void;
}

static Scheme_Object *
sch_exit (int argc, Scheme_Object *argv[])
{
  int status;

  if (argc > 1)
    scheme_wrong_count("exit", 0, 1, argc, argv);

  if (argc == 1) {
    if (!SCHEME_INTP(argv[0]))
      scheme_wrong_type("exit", "integer", 0, argc, argv);
    status = SCHEME_INT_VAL(argv[0]);
  } else
    status = 0;

  if (scheme_current_process->exit_handler) {
    Scheme_Object *p[1];

    p[0] = scheme_make_integer(status);
    scheme_apply(scheme_current_process->exit_handler, 1, p);
  } else if (scheme_exit)
    scheme_exit(status);
  else
    exit(status);

  return scheme_void;
}
