/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include "nummacs.h"
#include <math.h>
#include <string.h>
#include <ctype.h>
#ifndef DONT_IGNORE_FPE_SIGNAL
#include <signal.h>
#endif

#ifndef INEXACT_PRINT_DIGITS
#define INEXACT_PRINT_DIGITS "14"
#endif

/* globals */

/* locals */
static Scheme_Object *number_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *complex_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *real_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *rational_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *integer_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *exact_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *inexact_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *lt (int argc, Scheme_Object *argv[]);
static Scheme_Object *gt (int argc, Scheme_Object *argv[]);
static Scheme_Object *lt_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *gt_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *zero_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *positive_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *negative_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *add1 (int argc, Scheme_Object *argv[]);
static Scheme_Object *sub1 (int argc, Scheme_Object *argv[]);
static Scheme_Object *odd_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *even_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *sch_max (int argc, Scheme_Object *argv[]);
static Scheme_Object *sch_min (int argc, Scheme_Object *argv[]);
static Scheme_Object *plus (int argc, Scheme_Object *argv[]);
static Scheme_Object *minus (int argc, Scheme_Object *argv[]);
static Scheme_Object *mult (int argc, Scheme_Object *argv[]);
static Scheme_Object *div_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *abs_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *quotient (int argc, Scheme_Object *argv[]);
static Scheme_Object *rem_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *modulo (int argc, Scheme_Object *argv[]);
static Scheme_Object *binary_and (int argc, Scheme_Object *argv[]);
static Scheme_Object *binary_or (int argc, Scheme_Object *argv[]);
static Scheme_Object *binary_xor (int argc, Scheme_Object *argv[]);
static Scheme_Object *binary_not (int argc, Scheme_Object *argv[]);
static Scheme_Object *gcd (int argc, Scheme_Object *argv[]);
static Scheme_Object *lcm (int argc, Scheme_Object *argv[]);
static Scheme_Object *floor_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *ceiling (int argc, Scheme_Object *argv[]);
static Scheme_Object *truncate (int argc, Scheme_Object *argv[]);
static Scheme_Object *round (int argc, Scheme_Object *argv[]);
static Scheme_Object *exp_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *log_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *sin_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *cos_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *asin_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *acos_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *atan_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *sqrt_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *expt (int argc, Scheme_Object *argv[]);
static Scheme_Object *exact_to_inexact (int argc, Scheme_Object *argv[]);
static Scheme_Object *inexact_to_exact (int argc, Scheme_Object *argv[]);
static Scheme_Object *number_to_string (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_to_number (int argc, Scheme_Object *argv[]);

static Scheme_Object *random_seed(int argc, Scheme_Object *argv[]);
static Scheme_Object *sch_random(int argc, Scheme_Object *argv[]);

void
scheme_init_number (Scheme_Env *env)
{
  if (scheme_starting_up) {
#ifndef DONT_IGNORE_FPE_SIGNAL
    signal(SIGFPE, SIG_IGN);
#endif
  }

  scheme_add_global_constant ("number?", scheme_make_prim_w_arity(number_p, 1, 1), env);
  scheme_add_global_constant ("complex?", scheme_make_prim_w_arity(complex_p, 1, 1), env);
  scheme_add_global_constant ("real?", scheme_make_prim_w_arity(real_p, 1, 1), env);
  scheme_add_global_constant ("rational?", scheme_make_prim_w_arity(rational_p, 1, 1), env);
  scheme_add_global_constant ("integer?", scheme_make_prim_w_arity(integer_p, 1, 1), env);
  scheme_add_global_constant ("exact?", scheme_make_prim_w_arity(exact_p, 1, 1), env);
  scheme_add_global_constant ("inexact?", scheme_make_prim_w_arity(inexact_p, 1, 1), env);
  scheme_add_global_constant ("=", scheme_make_prim_w_arity(eq, 1, -1), env);
  scheme_add_global_constant ("<", scheme_make_prim_w_arity(lt, 1, -1), env);
  scheme_add_global_constant (">", scheme_make_prim_w_arity(gt, 1, -1), env);
  scheme_add_global_constant ("<=", scheme_make_prim_w_arity(lt_eq, 1, -1), env);
  scheme_add_global_constant (">=", scheme_make_prim_w_arity(gt_eq, 1, -1), env);
  scheme_add_global_constant ("zero?", scheme_make_prim_w_arity(zero_p, 1, 1), env);
  scheme_add_global_constant ("positive?", scheme_make_prim_w_arity(positive_p, 1, 1), env);
  scheme_add_global_constant ("negative?", scheme_make_prim_w_arity(negative_p, 1, 1), env);
  scheme_add_global_constant ("add1", scheme_make_prim_w_arity(add1, 1, 1), env);
  scheme_add_global_constant ("sub1", scheme_make_prim_w_arity(sub1, 1, 1), env);
  scheme_add_global_constant ("odd?", scheme_make_prim_w_arity(odd_p, 1, 1), env);
  scheme_add_global_constant ("even?", scheme_make_prim_w_arity(even_p, 1, 1), env);
  scheme_add_global_constant ("max", scheme_make_prim_w_arity(sch_max, 1, -1), env);
  scheme_add_global_constant ("min", scheme_make_prim_w_arity(sch_min, 1, -1), env);
  scheme_add_global_constant ("+", scheme_make_prim_w_arity(plus, 0, -1), env);
  scheme_add_global_constant ("-", scheme_make_prim_w_arity(minus, 1, -1), env);
  scheme_add_global_constant ("*", scheme_make_prim_w_arity(mult, 0, -1), env);
  scheme_add_global_constant ("/", scheme_make_prim_w_arity(div_prim, 1, -1), env);
  scheme_add_global_constant ("abs", scheme_make_prim_w_arity(abs_prim, 1, 1), env);
  scheme_add_global_constant ("quotient", scheme_make_prim_w_arity(quotient, 2, 2), env);
  scheme_add_global_constant ("remainder", scheme_make_prim_w_arity(rem_prim, 2, 2), env);
  scheme_add_global_constant ("modulo", scheme_make_prim_w_arity(modulo, 2, 2), env);
  scheme_add_global_constant ("binary-and", scheme_make_prim_w_arity(binary_and, 0, -1), env);
  scheme_add_global_constant ("binary-or", scheme_make_prim_w_arity(binary_or, 0, -1), env);
  scheme_add_global_constant ("binary-xor", scheme_make_prim_w_arity(binary_xor, 0, -1), env);
  scheme_add_global_constant ("binary-not", scheme_make_prim_w_arity(binary_not, 1, 1), env);
  scheme_add_global_constant ("gcd", scheme_make_prim_w_arity(gcd, 0, -1), env);
  scheme_add_global_constant ("lcm", scheme_make_prim_w_arity(lcm, 0, -1), env);
  scheme_add_global_constant ("floor", scheme_make_prim_w_arity(floor_prim, 1, 1), env);
  scheme_add_global_constant ("ceiling", scheme_make_prim_w_arity(ceiling, 1, 1), env);
  scheme_add_global_constant ("truncate", scheme_make_prim_w_arity(truncate, 1, 1), env);
  scheme_add_global_constant ("round", scheme_make_prim_w_arity(round, 1, 1), env);
  scheme_add_global_constant ("exp", scheme_make_prim_w_arity(exp_prim, 1, 1), env);
  scheme_add_global_constant ("log", scheme_make_prim_w_arity(log_prim, 1, 1), env);
  scheme_add_global_constant ("sin", scheme_make_prim_w_arity(sin_prim, 1, 1), env);
  scheme_add_global_constant ("cos", scheme_make_prim_w_arity(cos_prim, 1, 1), env);
  scheme_add_global_constant ("asin", scheme_make_prim_w_arity(asin_prim, 1, 2), env);
  scheme_add_global_constant ("acos", scheme_make_prim_w_arity(acos_prim, 1, 1), env);
  scheme_add_global_constant ("atan", scheme_make_prim_w_arity(atan_prim, 1, 1), env);
  scheme_add_global_constant ("sqrt", scheme_make_prim_w_arity(sqrt_prim, 1, 1), env);
  scheme_add_global_constant ("expt", scheme_make_prim_w_arity(expt, 2, 2), env);
  scheme_add_global_constant ("random", scheme_make_prim_w_arity(sch_random, 1, 1), env);
  scheme_add_global_constant ("random-seed", scheme_make_prim_w_arity(random_seed, 1, 1), env);
  scheme_add_global_constant ("exact->inexact", scheme_make_prim_w_arity(exact_to_inexact, 1, 1), env);
  scheme_add_global_constant ("inexact->exact", scheme_make_prim_w_arity(inexact_to_exact, 1, 1), env);
  scheme_add_global_constant ("number->string", scheme_make_prim_w_arity(number_to_string, 1, 2), env);
  scheme_add_global_constant ("string->number", scheme_make_prim_w_arity(string_to_number, 1, 2), env);
}


Scheme_Object *
scheme_make_integer_value (long i)
{
#ifndef FAST_NUMBERS
  Scheme_Object *si;

  si = scheme_alloc_stubborn_small_object ();
  si->type = scheme_integer_type;
  SCHEME_INT_VAL (si) = i;
  scheme_end_stubborn_change((void *)si);
  return (si);
#else
  return scheme_make_integer(i);
#endif
}

Scheme_Object *
scheme_make_double (double d)
{
  Scheme_Double *sd;

  sd = (Scheme_Double *)scheme_malloc_stubborn(sizeof(Scheme_Double));
  sd->type = scheme_double_type;
  SCHEME_DBL_VAL (sd) = d;
  scheme_end_stubborn_change((void *)sd);
  return (Scheme_Object *)sd;
}

/* locals */

#define TAKES_ONE(name) if (argc != 1) \
  scheme_wrong_count(#name, 1, 1, argc, argv)
#define NEED_NUMBER(name) \
  scheme_wrong_type(#name, "number", 0, argc, argv)
#define WRONG_TYPE(name, expected, value) \
  scheme_wrong_type(name, expected, -1, 0, &value)

static Scheme_Object *
number_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(number?);
  return (SCHEME_NUMBERP(argv[0]) ? scheme_true : scheme_false);
}

static Scheme_Object *
complex_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(complex?);
  return (SCHEME_NUMBERP(argv[0]) ? scheme_true : scheme_false);
}

static Scheme_Object *
real_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(real?);
  return (SCHEME_NUMBERP(argv[0]) ? scheme_true : scheme_false);
}

static Scheme_Object *
rational_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(rational?);

  return (SCHEME_INTP(argv[0]) || SCHEME_BIGNUMP(argv[0])
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
integer_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(integer?);

  if (SCHEME_INTP(argv[0]) || SCHEME_BIGNUMP(argv[0]))
    return scheme_true;

  if (SCHEME_DBLP(argv[0])) {
    double d;
    d = SCHEME_DBL_VAL(argv[0]);
    if (floor(d) == d)
      return scheme_true;
  }

  return scheme_false;
}

static Scheme_Object *
exact_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(exact?);
  return (SCHEME_INTP(argv[0]) || SCHEME_BIGNUMP(argv[0])
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
inexact_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(inexact?);
  return (SCHEME_DBLP(argv[0]) ? scheme_true : scheme_false);
}

GEN_BIN_COMP_PROT(bin_eq);
GEN_BIN_COMP_PROT(bin_lt);
GEN_BIN_COMP_PROT(bin_gt);
GEN_BIN_COMP_PROT(bin_lt_eq);
GEN_BIN_COMP_PROT(bin_gt_eq);

GEN_NARY_COMP(eq, "=", bin_eq)
GEN_NARY_COMP(lt, "<", bin_lt)
GEN_NARY_COMP(gt, ">", bin_gt)
GEN_NARY_COMP(lt_eq, "<=", bin_lt_eq)
GEN_NARY_COMP(gt_eq, ">=", bin_gt_eq)

#define EQUAL(x, y) (x == y)
#define LESS_THAN(x, y) (x < y)
#define GREATER_THAN(x, y) (x > y)
#define LESS_OR_EQUAL(x, y) (x <= y)
#define GREATER_OR_EQUAL(x, y) (x >= y)

GEN_BIN_COMP(bin_eq, "=", EQUAL, scheme_bignum_eq)
GEN_BIN_COMP(bin_lt, "<", LESS_THAN, scheme_bignum_lt)
GEN_BIN_COMP(bin_gt, ">", GREATER_THAN, scheme_bignum_gt)
GEN_BIN_COMP(bin_lt_eq, "<=", LESS_OR_EQUAL, scheme_bignum_le)
GEN_BIN_COMP(bin_gt_eq, ">=", GREATER_OR_EQUAL, scheme_bignum_ge)

static Scheme_Object *
zero_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(zero?);
  
  if (SCHEME_INTP(argv[0]))
#ifdef FAST_NUMBERS
    return (argv[0] == scheme_make_integer(0)) ? scheme_true : scheme_false;
#else
    return SCHEME_INT_VAL(argv[0]) ? scheme_false : scheme_true;
#endif
  if (SCHEME_DBLP(argv[0]))
    return SCHEME_DBL_VAL(argv[0]) ? scheme_false : scheme_true;
  if (SCHEME_BIGNUMP(argv[0]))
    return scheme_false;
 
  NEED_NUMBER(zero?);

  return scheme_false;
}

static Scheme_Object *
positive_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(positive?);

  if (SCHEME_INTP(argv[0]))
    return (SCHEME_INT_VAL(argv[0])>0 ? scheme_true : scheme_false);
  if (SCHEME_DBLP(argv[0]))
    return (SCHEME_DBL_VAL(argv[0])>0 ? scheme_true : scheme_false);
  if (SCHEME_BIGNUMP(argv[0]))
    return (SCHEME_PINT_VAL(argv[0])>0 ? scheme_true : scheme_false);

  NEED_NUMBER(positive?);

  return scheme_void;
}

static Scheme_Object *
negative_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(negative?);

  if (SCHEME_INTP(argv[0]))
    return (SCHEME_INT_VAL(argv[0])<0 ? scheme_true : scheme_false);
  if (SCHEME_DBLP(argv[0]))
    return (SCHEME_DBL_VAL(argv[0])<0 ? scheme_true : scheme_false);
  if (SCHEME_BIGNUMP(argv[0]))
    return (SCHEME_PINT_VAL(argv[0])<0 ? scheme_true : scheme_false);

  NEED_NUMBER(positive?);

  return scheme_void;
}

static Scheme_Object *
odd_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(odd?);

  if (SCHEME_INTP(argv[0]))
    return (((SCHEME_INT_VAL(argv[0])%2) != 0) ? scheme_true : scheme_false);
  if (SCHEME_DBLP(argv[0]))
    return (scheme_false);
  if (SCHEME_BIGNUMP(argv[0]))
    return (((bignum *)SCHEME_IPTR_VAL(argv[0]))[0] % 2) ? scheme_true : scheme_false;

  NEED_NUMBER(odd?);

  return scheme_void;
}

static Scheme_Object *
even_p (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(even?);

  if (SCHEME_INTP(argv[0]))
    return ((SCHEME_INT_VAL(argv[0])%2)==0 ? scheme_true : scheme_false);
  if (SCHEME_DBLP(argv[0]))
    return (scheme_false);
  if (SCHEME_BIGNUMP(argv[0]))
    return (((bignum *)SCHEME_IPTR_VAL(argv[0]))[0] % 2) ? scheme_false : scheme_true;

  NEED_NUMBER(even?);

  return scheme_void;
}

static Scheme_Object *
add1 (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(add1);

  if (SCHEME_INTP(argv[0])) {
    int v;
    v = SCHEME_INT_VAL(argv[0]);
    if (v < 0x3FFFFFFF)
      return scheme_make_integer(v + 1);
    else
      return scheme_bignum_add1(scheme_make_bignum(v));
  }
  if (SCHEME_DBLP(argv[0]))
    return scheme_make_double(SCHEME_DBL_VAL(argv[0]) + 1.0);
  if (SCHEME_BIGNUMP(argv[0]))
    return scheme_bignum_add1(argv[0]);

  NEED_NUMBER(add1);

  return scheme_void;
}

static Scheme_Object *
sub1 (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(sub1);

  if (SCHEME_INTP(argv[0])) {
    int v;
    v = SCHEME_INT_VAL(argv[0]);
    if (v > -(0x3FFFFFFF))
      return scheme_make_integer(SCHEME_INT_VAL(argv[0]) - 1);
    else
      return scheme_bignum_sub1(scheme_make_bignum(v));
  }
  if (SCHEME_DBLP(argv[0]))
    return scheme_make_double(SCHEME_DBL_VAL(argv[0]) - 1.0);
  if (SCHEME_BIGNUMP(argv[0]))
    return scheme_bignum_sub1(argv[0]);
  
  NEED_NUMBER(sub1);

  return scheme_void;
}

GEN_BIN_PROT(bin_plus);
GEN_BIN_PROT(bin_minus);
GEN_BIN_PROT(bin_mult);
GEN_BIN_PROT(bin_div);
GEN_BIN_PROT(bin_max);
GEN_BIN_PROT(bin_min);

#define F_ADD(x,y) scheme_make_double(x + y)
#define F_SUBTRACT(x,y) scheme_make_double(x - y)
#define F_MULTIPLY(x,y) scheme_make_double(x * y)
#define DIVIDE(x,y) scheme_make_double((double)x / (double)y)
#define MAX(n1,n2) scheme_make_integer((n1>n2) ? n1 : n2)
#define MIN(n1,n2) scheme_make_integer((n1<n2) ? n1 : n2)
#define F_MAX(n1,n2) scheme_make_double((n1>n2) ? n1 : n2)
#define F_MIN(n1,n2) scheme_make_double((n1<n2) ? n1 : n2)

static Scheme_Object *ADD(int a, int b)
{
  int r;
  Scheme_Object *o;

  r = a + b;

  o = scheme_make_integer(r);
  r = SCHEME_INT_VAL(o);

  if (b == r - a)
    return o;
  else {
    Small_Bignum sa, sb;
    return scheme_bignum_add(scheme_make_small_bignum(a, &sa),
			     scheme_make_small_bignum(b, &sb));
  }
}

static Scheme_Object *SUBTRACT(int a, int b)
{
  int r;
  Scheme_Object *o;

  r = a - b;

  o = scheme_make_integer(r);
  r = SCHEME_INT_VAL(o);

  if (a == r + b)
    return o;
  else {
    Small_Bignum sa, sb;
    return scheme_bignum_subtract(scheme_make_small_bignum(a, &sa),
				  scheme_make_small_bignum(b, &sb));
  }
}

static Scheme_Object *MULTIPLY(int a, int b)
{
  int r;
  Scheme_Object *o;

  if (!b)
    return scheme_make_integer(0);

  r = a * b;

  o = scheme_make_integer(r);
  r = SCHEME_INT_VAL(o);

  if (a == r / b)
    return o;
  else {
    Small_Bignum sa, sb;
    return scheme_bignum_multiply(scheme_make_small_bignum(a, &sa),
				  scheme_make_small_bignum(b, &sb));
  }
}

GEN_BIN_OP(bin_plus, "+", ADD, F_ADD, scheme_bignum_add, GEN_IDENT)
GEN_BIN_OP(bin_minus, "-", SUBTRACT, F_SUBTRACT, scheme_bignum_subtract, GEN_IDENT)
GEN_BIN_OP(bin_mult, "*", MULTIPLY, F_MULTIPLY, scheme_bignum_multiply, GEN_IDENT)
GEN_BIN_OP(bin_div, "/", DIVIDE, DIVIDE, DIVIDE, scheme_bignum_to_float)
GEN_BIN_OP(bin_max, "max", MAX, F_MAX, scheme_bignum_max, GEN_IDENT)
GEN_BIN_OP(bin_min, "min", MIN, F_MIN, scheme_bignum_min, GEN_IDENT)

GEN_BIN_PROT(bin_binary_and);
GEN_BIN_PROT(bin_binary_or);
GEN_BIN_PROT(bin_binary_xor);

GEN_BIN_INT_OP(bin_binary_and, "binary-and", &)
GEN_BIN_INT_OP(bin_binary_or, "binary-or", |)
GEN_BIN_INT_OP(bin_binary_xor, "binary-xor", ^)

GEN_NARY_OP(binary_and, "binary-and", bin_binary_and, -1)
GEN_NARY_OP(binary_or, "binary-or", bin_binary_or, 0)
GEN_NARY_OP(binary_xor, "binary-xor", bin_binary_xor, -1)

GEN_TWOARY_OP(sch_max, "max", bin_max)
GEN_TWOARY_OP(sch_min, "min", bin_min)
GEN_NARY_OP(plus, "+", bin_plus, 0)
GEN_NARY_OP(mult, "*", bin_mult, 1)

static Scheme_Object *
minus (int argc, Scheme_Object *argv[])
{
  Scheme_Object *ret;
  int i;

  if (!argc)
    scheme_wrong_count("-", 1, -1, argc, argv);

  ret = argv[0];
  if (argc == 1)
    ret = bin_minus (scheme_make_integer(0), ret);
  for ( i=1 ; i<argc ; ++i )
    ret = bin_minus (ret, argv[i]);
  return (ret);
}

static Scheme_Object *
div_prim (int argc, Scheme_Object *argv[])
{
  Scheme_Object *ret;
  int i;

  if (!argc)
    scheme_wrong_count("/", 1, -1, argc, argv);

  ret = argv[0];
  if (argc == 1)
    ret = bin_div (scheme_make_integer(1), ret);
  for ( i=1 ; i<argc ; ++i )
    ret = bin_div (ret, argv[i]);
  return (ret);
}

#define ABS(n)  ((n>0) ? n : -n)

static Scheme_Object *
abs_prim (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(abs);

  if (SCHEME_INTP(argv[0]))
    return (scheme_make_integer (ABS (SCHEME_INT_VAL(argv[0]))));
  if (SCHEME_DBLP(argv[0]))
    return (scheme_make_double (fabs (SCHEME_DBL_VAL(argv[0]))));
  if (SCHEME_BIGNUMP(argv[0])) {
    if (SCHEME_PINT_VAL(argv[0]) > 0)
      return argv[0];
    return scheme_bignum_negate(argv[0]);
  }

  NEED_NUMBER(abs);

  return scheme_void;
}

static Scheme_Object *bin_quotient (Scheme_Object *n1, Scheme_Object *n2);

static Scheme_Object *to_bignum(Scheme_Object *o)
{
  if (SCHEME_INTP(o))
    return scheme_make_bignum(SCHEME_INT_VAL(o));
  else if (SCHEME_DBLP(o))
    return scheme_bignum_from_float(SCHEME_DBL_VAL(o));
  else
    return o;
}

static Scheme_Object *
bin_quotient (Scheme_Object *n1, Scheme_Object *n2)
{
  Scheme_Object *q;

  if (!SCHEME_NUMBERP(n1)) WRONG_TYPE("quotient", "number", n1);
  if (!SCHEME_NUMBERP(n2)) WRONG_TYPE("quotient", "number", n2);

  if (SCHEME_INTP(n1) && SCHEME_INTP(n2))
    return (scheme_make_integer (SCHEME_INT_VAL(n1) / SCHEME_INT_VAL(n2)));
  if (SCHEME_DBLP(n1) || SCHEME_DBLP(n2))
    return bin_div(n1, n2);

  n1 = to_bignum(n1);
  n2 = to_bignum(n2);

  scheme_bignum_divide(n1, n2, &q, NULL, 1);
  return q;
}

static Scheme_Object *
quotient (int argc, Scheme_Object *argv[])
{
  if (argc != 2)
    scheme_wrong_count("quotient", 2, 2, argc, argv);

  return (bin_quotient (argv[0], argv[1]));
}

static Scheme_Object *
rem_mod (int argc, Scheme_Object *argv[], char *name, int first_sign)
{
  Scheme_Object *n1, *n2, *r;
  int negate;

  if (argc != 2)
    scheme_wrong_count(name, 2, 2, argc, argv);

  n1 = argv[0];
  n2 = argv[1];

  if (!SCHEME_NUMBERP(n1))
    scheme_wrong_type(name, "number", 0, argc, argv);
  if (!SCHEME_NUMBERP(n2))
    scheme_wrong_type(name, "number", 1, argc, argv);

  if (SCHEME_INTP(n1) && SCHEME_INTP(n2)) {
    int a, b, na, nb, v, neg1, neg2;

    a = SCHEME_INT_VAL(n1);
    b = SCHEME_INT_VAL(n2);
    na =  (a < 0) ? -a : a;
    nb =  (b < 0) ? -b : b;

    v = na % nb;

    if (first_sign) {
      if (a < 0)
	v = -v;
    } else {
      neg1 = (a < 0);
      neg2 = (b < 0);

      if (neg1 != neg2)
	v = nb - v;

      if (neg2)
	v = -v;
    }

    return scheme_make_integer(v);
  }

  if (SCHEME_DBLP(n1) || SCHEME_DBLP(n2)) {
    double a, b, na, nb, v;

    if (SCHEME_INTP(n1))
      a = SCHEME_INT_VAL(n1);
    else if (SCHEME_DBLP(n1))
      a = SCHEME_DBL_VAL(n1);
    else
      a = scheme_bignum_to_float(n1);

    if (SCHEME_INTP(n2))
      b = SCHEME_INT_VAL(n2);
    else if (SCHEME_DBLP(n2))
      b = SCHEME_DBL_VAL(n2);
    else
      b = scheme_bignum_to_float(n2);
    na =  (a < 0) ? -a : a;
    nb =  (b < 0) ? -b : b;

    v = fmod(na, nb);

    if (first_sign) {
      if (a < 0)
	v = -v;
    } else {
      int neg1, neg2;

      neg1 = (a < 0);
      neg2 = (b < 0);

      if (neg1 != neg2)
	v = nb - v;

      if (neg2)
	v = -v;
    }

    return scheme_make_double(v);
  }

  n1 = to_bignum(n1);
  n2 = to_bignum(n2);

  scheme_bignum_divide(n1, n2, NULL, &r, 1);

  negate = 0;

  if (!SCHEME_INTP(r) || SCHEME_INT_VAL(r)) {
    if (first_sign) {
      if (SCHEME_PINT_VAL(n1) < 0)
	negate = 1;
    } else {
      int neg1, neg2;
      
      neg1 = (SCHEME_PINT_VAL(n1) < 0);
      neg2 = (SCHEME_PINT_VAL(n2) < 0);
      
      if (neg1 != neg2) {
	if (neg2)
	  r = bin_plus(n2, r);
	else
	  r = bin_minus(n2, r);
      } else if (neg2)
	negate = 1;
    }
    
    if (negate) {
      if (SCHEME_INTP(r))
	r = scheme_make_integer(-SCHEME_INT_VAL(r));
      else
	r = scheme_bignum_negate(r);
    }
  }

  return r;
}

static Scheme_Object *
rem_prim (int argc, Scheme_Object *argv[])
{
  return rem_mod(argc, argv, "remainder", 1);
}

static Scheme_Object *
modulo (int argc, Scheme_Object *argv[])
{
  return rem_mod(argc, argv, "modulo", 0);
}

static Scheme_Object *bin_gcd (Scheme_Object *n1, Scheme_Object *n2);
static Scheme_Object *bin_lcm (Scheme_Object *n1, Scheme_Object *n2);

GEN_NARY_OP(gcd, gcd, bin_gcd, 0)
GEN_NARY_OP(lcm, lcm, bin_lcm, 1)

static Scheme_Object *
bin_gcd (Scheme_Object *n1, Scheme_Object *n2)
{
  if (!SCHEME_NUMBERP(n1)) WRONG_TYPE("gcd", "number", n1);
  if (!SCHEME_NUMBERP(n2)) WRONG_TYPE("gcd", "number", n2);

  if (SCHEME_INTP(n1) && SCHEME_INTP(n2)) {
    int i1, i2, a, b, r;

    i1 = SCHEME_INT_VAL(n1);
    i2 = SCHEME_INT_VAL(n2);
    if (i1 < 0)
      i1 = -i1;
    if (i2 < 0)
      i2 = -i2;
    if (i1 > i2) {
      a = i1;
      b = i2;
    } else {
      a = i2;
      b = i1;
    }
    
    r = 1;
    while ((b > 0) && (r > 0))
      {
	r = a % b;
	a = b;
	b = r;
      }
    return (scheme_make_integer(a));
  } else if (SCHEME_DBLP(n1) || SCHEME_DBLP(n2)) {
    double i1, i2, a, b, r;

    if (SCHEME_INTP(n1))
      i1 = SCHEME_INT_VAL(n1);
    else if (SCHEME_DBLP(n1))
      i1 = SCHEME_DBL_VAL(n1);
    else
      i1 = scheme_bignum_to_float(n1);

    if (SCHEME_INTP(n2))
      i2 = SCHEME_INT_VAL(n2);
    else if (SCHEME_DBLP(n2))
      i2 = SCHEME_DBL_VAL(n2);
    else
      i2 = scheme_bignum_to_float(n2);

    if (i1 < 0)
      i1 = -1;
    if (i2 < 0)
      i2 = -i2;
    if (i1 > i2) {
      a = i1;
      b = i2;
    } else {
      a = i2;
      b = i1;
    }
    
    r = 1;
    while ((b > 0) && (r > 0)) {
      r = fmod(a, b);
      a = b;
      b = r;
    }
    return (scheme_make_double(a));
  } else {
    Scheme_Object *save, *r;

    n1 = to_bignum(n1);
    n2 = to_bignum(n2);

    if (SCHEME_PINT_VAL(n1) < 0)
      n1 = scheme_bignum_negate(n1);
    if (SCHEME_PINT_VAL(n2) < 0)
      n2 = scheme_bignum_negate(n2);

    if (scheme_bignum_lt(n1, n2)) {
      save = n1;
      n1 = n2;
      n2 = save;
    }
    
    r = scheme_make_bignum(1);
    while (SCHEME_PINT_VAL(n2) && SCHEME_PINT_VAL(r)) {
      scheme_bignum_divide(n1, n2, NULL, &r, 0);
      n1 = n2;
      n2 = r;
    }

    return scheme_bignum_normalize(n1);
  }
}

static Scheme_Object *
bin_lcm (Scheme_Object *n1, Scheme_Object *n2)
{
  Scheme_Object *d, *ret;

  if (!SCHEME_NUMBERP(n1)) WRONG_TYPE("lcm", "number", n1);
  if (!SCHEME_NUMBERP(n2)) WRONG_TYPE("lcm", "number", n2);

  d = bin_gcd (n1, n2);
  
  if (NOT_SAME_OBJ(zero_p(1, &d), scheme_false))
    return d;

  ret = bin_mult (n1, bin_quotient (n2, d));

  return abs_prim(1, &ret);
}

static Scheme_Object *
floor_prim (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(floor);

  if (SCHEME_INTP (argv[0]))
    return (argv[0]);
  if (SCHEME_DBLP (argv[0]))
    return scheme_make_double(floor(SCHEME_DBL_VAL(argv[0])));
  if (SCHEME_BIGNUMP(argv[0]))
    return argv[0];

  NEED_NUMBER(floor);

  return scheme_void;
}

static Scheme_Object *
ceiling (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(ceiling);

  if (SCHEME_INTP (argv[0]))
    return (argv[0]);
  if (SCHEME_DBLP (argv[0]))
    return scheme_make_double(ceil(SCHEME_DBL_VAL(argv[0])));
  if (SCHEME_BIGNUMP(argv[0]))
    return argv[0];

  NEED_NUMBER(ceiling);

  return scheme_void;
}

static Scheme_Object *
truncate (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(truncate);

  if (SCHEME_INTP (argv[0]))
    return (argv[0]);
  if (SCHEME_DBLP (argv[0])) {
    double v = SCHEME_DBL_VAL(argv[0]);
    if (v > 0)
      v = floor(v);
    else
      v = ceil(v);
    return scheme_make_double(v);
  }
  if (SCHEME_BIGNUMP(argv[0]))
    return argv[0];

  NEED_NUMBER(truncate);

  return scheme_void;
}

static Scheme_Object *
round (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(round);

  if (SCHEME_INTP (argv[0]))
    return (argv[0]);
  if (SCHEME_DBLP (argv[0])) {
    double d = SCHEME_DBL_VAL(argv[0]);
    double i, frac;
    int invert;

    if (d < 0) {
      d = -d;
      invert = 1;
    } else
      invert = 0;

    frac = modf(d, &i);
    if (frac < 0.5)
      d = i;
    else if (frac > 0.5)
      d = i + 1;
    else if (fmod(i, 2.0))
	d = i + 1;
    else
      d = i;

    if (invert)
      d = -d;

    return scheme_make_double(d);
  } if (SCHEME_BIGNUMP(argv[0]))
    return argv[0];

  NEED_NUMBER(round);

  return scheme_void;
}

GEN_UNARY_OP(exp_prim, exp, exp)
GEN_UNARY_OP(log_prim, log, log)
GEN_UNARY_OP(sin_prim, sin, sin)
GEN_UNARY_OP(cos_prim, cos, cos)
GEN_UNARY_OP(asin_prim, asin, asin)
GEN_UNARY_OP(acos_prim, acos, acos)

static Scheme_Object *
atan_prim (int argc, Scheme_Object *argv[])
{
  double v;

  if (argc!=1 && argc!=2)
    scheme_wrong_count("atan", 1, 2, argc, argv);

  if (SCHEME_INTP (argv[0]))
    v = SCHEME_INT_VAL (argv[0]);
  else if (SCHEME_DBLP (argv[0]))
    v = SCHEME_DBL_VAL (argv[0]);
  else if (SCHEME_BIGNUMP (argv[0]))
    v = scheme_bignum_to_float(argv[0]);
  else {
    NEED_NUMBER(atan);
    v = 0.0;
  }

  if (argc == 2) {
    double v2;

    if (SCHEME_INTP (argv[1]))
      v2 = SCHEME_INT_VAL (argv[1]);
    else if (SCHEME_DBLP (argv[1]))
      v2 = SCHEME_DBL_VAL (argv[1]);
    else if (SCHEME_BIGNUMP (argv[1]))
      v2 = scheme_bignum_to_float(argv[1]);
    else {
      scheme_wrong_type("atan", "number", 1, argc, argv);
      v2 = 0.0;
    }

    v = atan2(v, v2);
  } else
    v = atan(v);

  return scheme_make_double(v);
}

static Scheme_Object *
sqrt_prim (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(sqrt);

  if (SCHEME_INTP (argv[0]))
    return (scheme_make_double (sqrt (SCHEME_INT_VAL (argv[0]))));
  if (SCHEME_DBLP (argv[0]))
    return (scheme_make_double (sqrt (SCHEME_DBL_VAL (argv[0]))));
  if (SCHEME_BIGNUMP (argv[0]))
    return (scheme_make_double (sqrt (scheme_bignum_to_float(argv[0]))));

  NEED_NUMBER(sqrt);

  return scheme_void;
}

#define EXPT(x, y) scheme_make_double(pow(x, y))

GEN_BIN_OP(bin_expt, "expt", EXPT, EXPT, EXPT, scheme_bignum_to_float)

static Scheme_Object *
expt (int argc, Scheme_Object *argv[])
{
  Scheme_Object *n1, *n2;
  if (argc != 2)
    scheme_wrong_count("expt", 2, 2, argc, argv);
  if (!SCHEME_NUMBERP(argv[0]))
    scheme_wrong_type("expt", "number", 0, argc, argv);
  if (!SCHEME_NUMBERP(argv[1]))
    scheme_wrong_type("expt", "number", 1, argc, argv);

  n1 = argv[0]; n2 = argv[1];

  if (SCHEME_DBLP(n1) || SCHEME_DBLP(n2) 
      || (SCHEME_INTP(n2) && SCHEME_INT_VAL(n2) < 0)
      || (SCHEME_BIGNUMP(n2) && SCHEME_PINT_VAL(n2) < 0))
    return bin_expt(n1, n2);

  n1 = to_bignum(n1);
  n2 = to_bignum(n2);
  return scheme_bignum_power(n1, n2);
}


static Scheme_Object *
exact_to_inexact (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(exact->inexact);

  if (SCHEME_INTP (argv[0]))
    return (scheme_make_double (SCHEME_INT_VAL(argv[0])));
  if (SCHEME_DBLP (argv[0]))
    return (argv[0]);
  if (SCHEME_BIGNUMP (argv[0]))
    return scheme_make_double(scheme_bignum_to_float(argv[0]));

  NEED_NUMBER(exact->inexact);

  return scheme_void;
}

static Scheme_Object *
inexact_to_exact (int argc, Scheme_Object *argv[])
{
  TAKES_ONE(inexact->exact);

  if (SCHEME_INTP (argv[0]))
    return (argv[0]);
  if (SCHEME_DBLP (argv[0]))
    return (scheme_bignum_from_float(SCHEME_DBL_VAL (argv[0])));
  if (SCHEME_BIGNUMP (argv[0]))
    return argv[0];

  NEED_NUMBER(inexact->exact);

  return scheme_void;
}

static Scheme_Object *
number_to_string (int argc, Scheme_Object *argv[])
{
  int radix;
  Scheme_Object *n;

  if (argc != 1 && argc != 2)
    scheme_wrong_count("number->string", 1, 2, argc, argv);

  if (argc == 2) {
    if (!SCHEME_INTP(argv[1]))
      scheme_wrong_type("number->string", "integer", 1, argc, argv);

    radix = SCHEME_INT_VAL(argv[1]);
 } else
   radix = 10;

  if (SCHEME_DBLP(argv[0])) {
    if (radix != 10)
      scheme_signal_error("number->string: "
			  "floats can only be printed in base 10");
    return (scheme_double_to_string (SCHEME_DBL_VAL(argv[0])));
  }
  
  if (!SCHEME_INTP(argv[0]) && !SCHEME_BIGNUMP(argv[0]))
    scheme_wrong_type("number->string", "number", 0, argc, argv);

  if (SCHEME_INTP(argv[0]))
    n = scheme_make_bignum(SCHEME_INT_VAL(argv[0]));
  else
    n = argv[0];

  return scheme_make_string_without_copying(scheme_bignum_to_string(n, radix));
}

Scheme_Object *scheme_double_to_string (double d)
{
  char buffer[100];
  int l, i;

  sprintf (buffer, "%." INEXACT_PRINT_DIGITS "g", d);
  l = strlen(buffer);
  for (i = 0; i < l; i++)
    if (buffer[i] == '.' || isalpha(buffer[i]))
      break;
  if (i == l) {
    buffer[i] = '.';
    buffer[i + 1] = '0';
    buffer[i + 2] = 0;
    l += 2;
  }
  return (scheme_make_string (buffer));
}

static Scheme_Object *
string_to_number (int argc, Scheme_Object *argv[])
{
  int radix, set;
  char *str;

  SCHEME_ASSERT ((argc == 1 || argc == 2), "string->number: wrong number of args");
  SCHEME_ASSERT (SCHEME_STRINGP(argv[0]), "string->number: first arg must be a string");

  if (argc == 2) {
    SCHEME_ASSERT(SCHEME_INTP(argv[1]), 
		   "string->number: second arg must be an integer");
    radix = SCHEME_INT_VAL (argv[1]);
    set = 1;
  } else {
    radix = 10;
    set = 0;
  }

  str = SCHEME_STR_VAL(argv[0]);

  return scheme_read_number(str, 0, 0, radix, set, 0);
}

Scheme_Object *scheme_read_number(const char *str, int is_float, 
				  int is_not_float,
				  int radix, int radix_set, int complain)
{
  double d;
  int len, i, has_decimal, must_parse, has_slash;
  Scheme_Object *o;
  char *ptr;

  while (isspace(*str))
    str++;

  while (str[0] == '#') {
    if (str[1] != 'e' && str[1] != 'i') {
      if (radix_set) {
	if (complain)
	  scheme_signal_error("read-number: multiple radix values given");
	else
	  return scheme_false;
      }
      radix_set = 1;
    } else {
      if (is_float || is_not_float) {
	if (complain)
	  scheme_signal_error("read-number: number %s can't be exact and inexact", str + 2);
	else
	  return scheme_false;
      }
    }

    switch (str[1]) {
    case 'b':
      radix = 2;
      break;
    case 'o':
      radix = 8;
      break;
    case 'd':
      radix = 10;
      break;
    case 'x':
      radix = 16;
      break;
    case 'i':
      is_float = 1;
      break;
    case 'e':
      is_not_float = 1;
      break;
    default:
      if (complain)
	scheme_signal_error("read-number: bad `#' indicator in %s", str);
      return scheme_false;
    }
    str += 2;
  }

  must_parse = (radix_set || is_float || is_not_float);

  len = strlen(str);
  if (!len) {
    if (complain && must_parse)
      scheme_signal_error("read-number: no digits");
    return scheme_false;
  }

  has_slash = 0;
  for (i= 0; i < len; i++) {
    int ch = str[i];
    if (ch == '/') {
      if (has_slash)
	 scheme_signal_error("read-number: only one slash allowed in number: "
			     "%s", str);
      if (!i) {
	if (complain && must_parse)
	  scheme_signal_error("read-number: slash without preceding digits: "
			      "%s", str);
	return scheme_false;
      }

      has_slash = i;
    }
  }

  if (has_slash) {
    Scheme_Object *n1, *n2;
    char *first;

    first = (char *)scheme_malloc_atomic(has_slash + 1);
    memcpy(first, str, has_slash);
    first[has_slash] = 0;

    n1 = scheme_read_number(first, 1, 0,
			    radix, 1, complain);
    n2 = scheme_read_number(str + has_slash + 1, 1, 0, 
			    radix, 1, complain);

    if (SAME_OBJ(n1, scheme_false) || SAME_OBJ(n2, scheme_false))
      return scheme_false;

    n1 = bin_div(n1, n2);

    if (is_not_float)
      return scheme_bignum_from_float(SCHEME_DBL_VAL(n1));
    else
      return n1;
  }

  has_decimal = 0;
  for (i= 0; i < len; i++) {
    int ch = str[i];
    if ((ch == '.') || (ch == 'e') || (ch == 'E')) {
      if (is_not_float) {
	if (complain)
	  scheme_signal_error("read-number: no exact fractions for %s", str);
	else
	  return scheme_false;
      }
      if (radix != 10) {
	if (complain)
	  scheme_signal_error("read-number: decimal implies base 10 in %s", 
			      str);
	return scheme_false;
      }
      has_decimal = 1;
      break;
    }
  }

  if (has_decimal) {
    d = strtod (str, &ptr);
    if ((ptr - str) < len) {
      if (complain && must_parse) {
	scheme_signal_error("read-number: bad decimal number %s", str);
      }
      return scheme_false;
    } 

    if (is_not_float)
      return scheme_bignum_from_float(d);
    else
      return scheme_make_double(d);
  }

  o = scheme_read_bignum(str, radix);
  if (SAME_OBJ(o, scheme_false)) {
    if (complain && must_parse)
      scheme_signal_error("read-number: bad number %s", str);
  } else if (is_float)
    return exact_to_inexact(1, &o);

  return o;
}

static Scheme_Object *
binary_not(int argc, Scheme_Object *argv[])
{
  TAKES_ONE(binary-not);
  if (!SCHEME_INTP(argv[0]))
    scheme_wrong_type("binary-not", "integer", 0, argc, argv);

  return scheme_make_integer(~SCHEME_INT_VAL(argv[0]));
}

static Scheme_Object *
random_seed(int argc, Scheme_Object *argv[])
{
  TAKES_ONE(random-seed);
  if (!SCHEME_INTP(argv[0]))
    scheme_wrong_type("random-seed", "integer", 0, argc, argv);

#ifdef RAND_NOT_RANDOM
  srand(SCHEME_INT_VAL(argv[0]));
#else
  srandom(SCHEME_INT_VAL(argv[0]));
#endif


  return scheme_void;
}

static Scheme_Object *
sch_random(int argc, Scheme_Object *argv[])
{
  int v;

  TAKES_ONE(random);
  if (!SCHEME_INTP(argv[0]))
    scheme_wrong_type("random", "integer", 0, argc, argv);

#ifdef RAND_NOT_RANDOM
  v = rand() % SCHEME_INT_VAL(argv[0]);
#else
  v = random();
  if (v < 0)
    v = -v;
  v = v % SCHEME_INT_VAL(argv[0]);
#endif

  return scheme_make_integer(v);
}
