/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "scheme.h"
#include "schpriv.h"
#include <ctype.h>

/* locals */
static char *print_buffer = NULL;
static long print_position;
static long print_allocated = 0;

static void print_to_port(char *name, Scheme_Object *obj, Scheme_Object *port, int escaped);
static int print(Scheme_Object *obj, int escaped, int compact);
static void print_string(Scheme_Object *string, int escaped);
static void print_pair(Scheme_Object *pair, int escaped, int compact);
static void print_vector (Scheme_Object *vec, int escaped, int compact);
static void print_char (Scheme_Object *chobj, int escaped);

void
scheme_debug_print (Scheme_Object *obj)
{
  scheme_write(obj, scheme_orig_stdout_port);
  fflush (stdout);
}

void
scheme_write(Scheme_Object *obj, Scheme_Object *port)
{
  print_to_port("write", obj, port, 1);
}

void
scheme_display(Scheme_Object *obj, Scheme_Object *port)
{
  print_to_port("display", obj, port, 0);
}

char *
scheme_print_to_string(Scheme_Object *obj, long *len, int write)
{
  long save_a, save_p;
  char *save_b, *newb;

  /* Printing is single-threaded right now, but that could change... */
  /* Better to handle it now and not forget. */
  save_a = print_allocated;
  save_p = print_position;
  save_b = print_buffer;

  print_allocated = 50;
  print_buffer = (char *)scheme_malloc_atomic(print_allocated);
  print_position = 0;

  print(obj, write, 0);

  print_buffer[print_position] = '\0';
  newb = print_buffer;

  if (len)
    *len = print_position;

  print_allocated = save_a;
  print_position = save_p;
  print_buffer = save_b;

  return newb;
}

char *scheme_write_to_string(Scheme_Object *obj, long *len)
{
  return scheme_print_to_string(obj, len, 1);
}

char *scheme_display_to_string(Scheme_Object *obj, long *len)
{
  return scheme_print_to_string(obj, len, 0);
}

void scheme_write_string(const char *str, long len, Scheme_Object *port)
{
  Scheme_Output_Port *op;
  op = (Scheme_Output_Port *)port;
  (op->write_string_fun)((char *)str, len, op);
}

static void 
print_to_port(char *name, Scheme_Object *obj, Scheme_Object *port, int escaped)
{
  Scheme_Output_Port *op;
  char *str;
  long len;
  
  op = (Scheme_Output_Port *)port;
  if (op->closed)
    scheme_signal_error("%s: port is closed", name);

  str = scheme_print_to_string(obj, &len, escaped);

  (op->write_string_fun)(str, len, op);
}

static void print_this_string(char *str, int autolen)
{
  long len;
  char *oldstr;

  if (autolen > 0)
    len = autolen;
  else
    len = strlen(str);

  if (len + print_position + 1 > print_allocated) {
    if (len + 1 >= print_allocated)
      print_allocated = 2 * print_allocated + len + 1;
    else
      print_allocated = 2 * print_allocated;

    oldstr = print_buffer;
    print_buffer = (char *)scheme_malloc_atomic(print_allocated);
    memcpy(print_buffer, oldstr, print_position);
  }

  memcpy(print_buffer + print_position, str, len);
  print_position += len;
}

static int
print (Scheme_Object *obj, int escaped, int compact)
{
  static char *buffer = NULL;
  int closed = 0;

  if (!buffer)
    buffer = (char *)scheme_malloc_atomic(100);

  if (SCHEME_SYMBOLP(obj))
    {
      print_this_string(SCHEME_SYM_VAL(obj), 0);
    }
  else if (SCHEME_TSYMBOLP(obj))
    { 
      print_this_string("#<", 2);
      print_this_string(SCHEME_SYM_VAL((Scheme_Object *)SCHEME_PTR_VAL(obj)), 0);
      print_this_string(">", 1);
    }
  else if (SCHEME_STRINGP(obj))
    {
      print_string (obj, escaped);
    }
  else if (SCHEME_CHARP(obj))
    {
      print_char (obj, escaped);
    }
  else if (SCHEME_INTP(obj))
    {
      sprintf (buffer, "%ld", SCHEME_INT_VAL (obj));
      print_this_string(buffer, 0);
    }
  else if (SCHEME_DBLP(obj))
    {
      Scheme_Object *str;

      str = scheme_double_to_string(SCHEME_DBL_VAL (obj));

      print_this_string(SCHEME_STR_VAL(str), 0);
    }
  else if (SCHEME_BIGNUMP(obj))
    {
      print_this_string(scheme_bignum_to_string(obj, 10), 0);
    }
  else if (SCHEME_NULLP(obj))
    {
      print_this_string("()", 2);
      closed = 1;
    }
  else if (SCHEME_PAIRP(obj))
    {
      print_pair(obj, escaped, compact);
      closed = 1;
    }
  else if (SCHEME_VECTORP(obj))
    {
      print_vector(obj, escaped, compact);
      closed = 1;
    }
  else if (SCHEME_BOXP(obj))
    {
      print_this_string("#&", 2);
      closed = print(SCHEME_PTR_VAL(obj), escaped, compact);
    }
  else if (SAME_OBJ(obj, scheme_true))
    {
      print_this_string("#t", 2);
    }
  else if (SAME_OBJ(obj, scheme_false))
    {
      print_this_string("#f", 2);
    }
  else if (SAME_TYPE(SCHEME_TYPE(obj), scheme_structure_type))
    {
      print_this_string("#<", 2);
      print_this_string(((Scheme_Structure *)obj)->type_name, 0);
      print_this_string(">", 1);
      closed = 1;
    }
  else if (scheme_type_writers[SCHEME_TYPE(obj)])
    {
      Scheme_Type t = SCHEME_TYPE(obj);
      
      print_this_string("#`", 2);
      if (t < _scheme_last_type_) {
	sprintf (buffer, "%ld", (long)SCHEME_TYPE(obj));
	print_this_string(buffer, 0);
      } else
	print_this_string(scheme_get_type_name(t), 0);
      closed = print((scheme_type_writers[t])(obj), escaped, 1);
    } 
  else
    {
      print_this_string("#", 1);
      print_this_string(scheme_get_type_name((SCHEME_TYPE(obj))), 0);
      closed = 1;
    }

  return closed;
}

static void
print_string (Scheme_Object *string, int escaped)
{
  char *str, minibuf[2];
  int simple;
  int len, i;

  if ( escaped )
    print_this_string("\"", 1);

  simple = 1;

  len = SCHEME_STRTAG_VAL(string);

  if (len) {
    if (escaped) {
      str = SCHEME_STR_VAL(string);
      for (i = 0; i < len; i++, str++)
	if ((*str == '"') || (*str == '\\') || (*str == '\0')) {
	  simple = 0;
	  break;
	}
    }
    
    if (simple)
      print_this_string(SCHEME_STR_VAL(string), len);
    else {
      minibuf[1] = 0;
      for (str = SCHEME_STR_VAL(string), i = 0; i < len; i++, str++) {
	if (!*str) {
	  print_this_string("\\0", 2);
	} else {
	  if ((*str == '"') || (*str == '\\'))
	    print_this_string("\\", 1);
	  minibuf[0] = *str;
	  print_this_string(minibuf, 1);
	}
      }
    }
  }

  if ( escaped )
    print_this_string("\"", 1);
}

static void
print_pair (Scheme_Object *pair, int escaped, int compact)
{
  Scheme_Object *cdr;
  int no_space_ok;

  print_this_string("(", 1);

  no_space_ok = print(SCHEME_CAR (pair), escaped, compact);

  cdr = SCHEME_CDR (pair);
  while (!SCHEME_NULLP(cdr) && SCHEME_PAIRP(cdr))
    {
      if (!compact || !no_space_ok)
	print_this_string(" ", 1);
      no_space_ok = print(SCHEME_CAR (cdr), escaped, compact);
      cdr = SCHEME_CDR (cdr);
    }

  if (!SCHEME_NULLP(cdr))
    {
      print_this_string(" . ", 3);
      print(cdr, escaped, compact);
    }
  print_this_string(")", 1);
}

static void
print_vector (Scheme_Object *vec, int escaped, int compact)
{
  int i, no_space_ok;

  print_this_string("#(", 2);
  for ( i=0 ; i<SCHEME_VEC_SIZE(vec) ; ++i )
    {
      no_space_ok = print(SCHEME_VEC_ELS(vec)[i], escaped, compact);
      if (i < SCHEME_VEC_SIZE(vec)-1)
	if (!compact || !no_space_ok)
	  print_this_string(" ", 1);
    }
  print_this_string(")", 1);
}

static void
print_char (Scheme_Object *charobj, int escaped)
{
  char ch, minibuf[4], *str;

  ch = SCHEME_CHAR_VAL (charobj);
  if (escaped)
    {
      switch ( ch )
	{
	case '\0':
	  str = "#\\null";
	  break;
	case '\n':
	  str = "#\\newline";
	  break;
	case '\t':
	  str = "#\\tab";
	  break;
	case ' ':
	  str  = "#\\space";
	  break;
	case '\r':
	  str = "#\\return";
	  break;
	case '\f':
	  str = "#\\page";
	  break;
	case '\b':
	  str = "#\\backspace";
	  break;
	default:
	  minibuf[0] = '#';
	  minibuf[1] = '\\';
	  minibuf[2] = ch;
	  minibuf[3] = 0;
	  str = minibuf;
	  break;
	}
    }
  else
    {
      minibuf[0] = ch;
      minibuf[1] = 0;
      str = minibuf;
    }

  print_this_string(str, 0);
}
