/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <string.h>

#include "schvers.h"

/* globals */

/* locals */
static Scheme_Object *string_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *make_string (int argc, Scheme_Object *argv[]);
static Scheme_Object *string (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_length (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_ref (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_set (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_ci_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_lt (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_gt (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_lt_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_gt_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_ci_lt (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_ci_gt (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_ci_lt_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_ci_gt_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *substring (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_append (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_to_list (int argc, Scheme_Object *argv[]);
static Scheme_Object *list_to_string (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_copy (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_fill (int argc, Scheme_Object *argv[]);
static Scheme_Object *version (int argc, Scheme_Object *argv[]);
static Scheme_Object *banner (int argc, Scheme_Object *argv[]);

#ifdef GETENV_FUNCTION
static Scheme_Object *sch_getenv(int argc, Scheme_Object *argv[]);
#endif

static int strcmp_ci (char *str1, char *str2);

void
scheme_init_string (Scheme_Env *env)
{
  scheme_add_global_constant ("string?", scheme_make_prim_w_arity (string_p, 1, 1), env);
  scheme_add_global_constant ("make-string", scheme_make_prim_w_arity (make_string, 1, 2), env);
  scheme_add_global_constant ("string", scheme_make_prim_w_arity (string, 0, -1), env);
  scheme_add_global_constant ("string-length", scheme_make_prim_w_arity (string_length, 1, 1), env);
  scheme_add_global_constant ("string-ref", scheme_make_prim_w_arity (string_ref, 2, 2), env);
  scheme_add_global_constant ("string-set!", scheme_make_prim_w_arity (string_set, 2, 2), env);
  scheme_add_global_constant ("string=?", scheme_make_prim_w_arity (string_eq, 2, 2), env);
  scheme_add_global_constant ("string-ci=?", scheme_make_prim_w_arity (string_ci_eq, 2, 2), env);
  scheme_add_global_constant ("string<?", scheme_make_prim_w_arity (string_lt, 2, 2), env);
  scheme_add_global_constant ("string>?", scheme_make_prim_w_arity (string_gt, 2, 2), env);
  scheme_add_global_constant ("string<=?", scheme_make_prim_w_arity (string_lt_eq, 2, 2), env);
  scheme_add_global_constant ("string>=?", scheme_make_prim_w_arity (string_gt_eq, 2, 2), env);
  scheme_add_global_constant ("string-ci<?", scheme_make_prim_w_arity (string_ci_lt, 2, 2), env);
  scheme_add_global_constant ("string-ci>?", scheme_make_prim_w_arity (string_ci_gt, 2, 2), env);
  scheme_add_global_constant ("string-ci<=?", scheme_make_prim_w_arity (string_ci_lt_eq, 2, 2), env);
  scheme_add_global_constant ("string-ci>=?", scheme_make_prim_w_arity (string_ci_gt_eq, 2, 2), env);
  scheme_add_global_constant ("substring", scheme_make_prim_w_arity (substring, 3, 3), env);
  scheme_add_global_constant ("string-append", scheme_make_prim_w_arity (string_append, 0, -1), env);
  scheme_add_global_constant ("string->list", scheme_make_prim_w_arity (string_to_list, 1, 1), env);
  scheme_add_global_constant ("list->string", scheme_make_prim_w_arity (list_to_string, 1, 1), env);
  scheme_add_global_constant ("string-copy", scheme_make_prim_w_arity (string_copy, 1, 1), env);
  scheme_add_global_constant ("string-fill!", scheme_make_prim_w_arity (string_fill, 2, 2), env);

  scheme_add_global_constant ("version", scheme_make_prim_w_arity(version, 0, 0), env);
  scheme_add_global_constant ("banner", scheme_make_prim_w_arity(banner, 0, 0), env);

#ifdef GETENV_FUNCTION
  scheme_add_global_constant ("getenv", scheme_make_prim_w_arity(sch_getenv, 1, 1), env);
#endif
}

Scheme_Object *
scheme_make_sized_string(char *chars, long len, int copy)
{
  Scheme_Object *str;
  
  str = scheme_alloc_stubborn_object ();
  str->type = scheme_string_type;
  if (len < 0)
    len = strlen(chars);
  if (copy) {
    SCHEME_STR_VAL(str) = scheme_malloc_atomic(len + 1);
    memcpy(SCHEME_STR_VAL(str), chars, len + 1);
  } else
    SCHEME_STR_VAL(str) = chars;
  SCHEME_STRTAG_VAL(str) = len;
  scheme_end_stubborn_change((void *)str);
  return (str);
}

Scheme_Object *
scheme_make_string_without_copying(char *chars)
{
  return scheme_make_sized_string(chars, -1, 0);
}

Scheme_Object *
scheme_make_string(const char *chars)
{
  return scheme_make_string_without_copying(scheme_strdup(chars));
}

Scheme_Object *
scheme_alloc_string (int size, char fill)
{
  Scheme_Object *str;
  int i;
  
  if (size < 0)
    scheme_signal_error("make-string: cannot make negative size");

  str = scheme_alloc_stubborn_object ();
  str->type = scheme_string_type;
  SCHEME_STR_VAL (str) = (char *) scheme_malloc_atomic (sizeof (char) * (size+1));
  for ( i=0 ; i<size ; ++i )
    {
      SCHEME_STR_VAL(str)[i] = fill;
    }
  SCHEME_STR_VAL(str)[i] = '\0';
  SCHEME_STRTAG_VAL(str) = size;

  scheme_end_stubborn_change((void *)str);

  return (str);
}

/* locals */

static Scheme_Object *
string_p (int argc, Scheme_Object *argv[])
{
  SCHEME_ASSERT ((argc == 1), "string?: wrong number of args");
  return (SCHEME_STRINGP(argv[0]) ? scheme_true : scheme_false);
}

static Scheme_Object *
make_string (int argc, Scheme_Object *argv[])
{
  int len;
  char fill;
  Scheme_Object *str;

  if ((argc != 1) && (argc != 2))
    scheme_wrong_count("make-string", 1, 2, argc, argv);
  if (!SCHEME_INTP(argv[0]) && !SCHEME_DBLP(argv[0]))
    scheme_wrong_type("make-string", "numberer", 0, argc, argv);

  len = SCHEME_NUM_VAL(argv[0]);

  if (argc == 2) {
    if (!SCHEME_CHARP(argv[1]))
      scheme_wrong_type("make-string", "character", 1, argc, argv);
    fill = SCHEME_CHAR_VAL (argv[1]);
  }
  else
    fill = ' ';

  str = scheme_alloc_string (len, fill);
  return (str);
}

static Scheme_Object *
string (int argc, Scheme_Object *argv[])
{
  Scheme_Object *str;
  int i;

  str = scheme_alloc_string (argc, 0);

  for ( i=0 ; i<argc ; ++i ) {
    if (!SCHEME_CHARP (argv[i]))
      scheme_wrong_type("string", "character", i, argc, argv);
    SCHEME_STR_VAL(str)[i] = SCHEME_CHAR_VAL(argv[i]);
  }

  return (str);
}

static Scheme_Object *
string_length (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("string-length", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("string-length", "string", 0, argc, argv);

  return scheme_make_integer(SCHEME_STRTAG_VAL(argv[0]));
}

static Scheme_Object *
string_ref (int argc, Scheme_Object *argv[])
{
  int i, len;
  char *str;

  if (argc != 2)
    scheme_wrong_count("string-ref", 2, 2, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("string-ref", "string", 0, argc, argv);
  if (!SCHEME_INTP(argv[1]) && !SCHEME_DBLP(argv[1]))
    scheme_wrong_type("string-ref", "number", 1, argc, argv);

  i= SCHEME_NUM_VAL(argv[1]);

  str = SCHEME_STR_VAL(argv[0]);
  len = SCHEME_STRTAG_VAL(argv[0]);

  if ((i < 0) || (i >= len))
    scheme_signal_error("string-ref: index %d out of range [%d, %d]", i,
			0, len - 1);

  return scheme_make_char(str[i]);
}

static Scheme_Object *
string_set (int argc, Scheme_Object *argv[])
{
  int i, len;
  char *str;

  if (argc != 3)
    scheme_wrong_count("string-set!", 3, 3, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("string-set!", "string", 0, argc, argv);
  if (!SCHEME_INTP(argv[1]) && !SCHEME_DBLP(argv[1]))
    scheme_wrong_type("string-set!", "number", 1, argc, argv);
  if (!SCHEME_CHARP(argv[2]))
    scheme_wrong_type("string-set!", "character", 2, argc, argv);

  str = SCHEME_STR_VAL(argv[0]);
  len = SCHEME_STRTAG_VAL(argv[0]);

  i= SCHEME_NUM_VAL(argv[1]);

  if ((i < 0) || (i >= len))
    scheme_signal_error("string-set!: index %d out of range [%d, %d]", i,
			0, len - 1);

  str[i] = SCHEME_CHAR_VAL (argv[2]);

  return scheme_void;
}

/* comparisons */

#define GEN_STRING_COMP(name, scheme_name, comp, op) \
static Scheme_Object * \
name (int argc, Scheme_Object *argv[]) \
{ \
   if (argc != 2) scheme_wrong_count(#scheme_name, 2, 2, argc, argv); \
   if (!SCHEME_STRINGP(argv[0])) \
    scheme_wrong_type(#scheme_name, "string", 0, argc, argv); \
   if (!SCHEME_STRINGP(argv[1])) \
    scheme_wrong_type(#scheme_name, "string", 1, argc, argv); \
  return ((comp (SCHEME_STR_VAL(argv[0]), SCHEME_STR_VAL(argv[1])) op 0) \
	  ? scheme_true : scheme_false); \
}

GEN_STRING_COMP(string_eq, "string=?", strcmp, ==)
GEN_STRING_COMP(string_ci_eq, "string-ci=?", strcmp_ci, ==)
GEN_STRING_COMP(string_lt, "string<?", strcmp, <)
GEN_STRING_COMP(string_gt, "string>?", strcmp, >)
GEN_STRING_COMP(string_lt_eq, "string<=?", strcmp, <=)
GEN_STRING_COMP(string_gt_eq, "string>=?", strcmp, >=)
GEN_STRING_COMP(string_ci_lt, "string-ci<?", strcmp_ci, <)
GEN_STRING_COMP(string_ci_gt, "string-ci>?", strcmp_ci, >)
GEN_STRING_COMP(string_ci_lt_eq, "string-ci<=?", strcmp_ci, <=)
GEN_STRING_COMP(string_ci_gt_eq, "string-ci>=?", strcmp_ci, >=)

static Scheme_Object *
substring (int argc, Scheme_Object *argv[])
{
  int len, start, finish, i;
  char *chars;
  Scheme_Object *str;

  if (argc != 3)
    scheme_wrong_count("substring", 3, 3, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("substring", "string", 0, argc, argv);
  if (!SCHEME_INTP(argv[1]) && !SCHEME_DBLP(argv[1]))
    scheme_wrong_type("substring", "number", 1, argc, argv);
  if (!SCHEME_INTP(argv[2]) && !SCHEME_DBLP(argv[2]))
    scheme_wrong_type("substring", "number", 2, argc, argv);

  chars = SCHEME_STR_VAL (argv[0]);
  len = SCHEME_STRTAG_VAL (argv[0]);
  start = SCHEME_NUM_VAL(argv[1]);
  finish = SCHEME_NUM_VAL(argv[2]);

  if (!(start >= 0 && start <= len))
    scheme_signal_error("substring: first index %d out of range [%d, %d]",
			start , 0, len);
  if (!(finish >= start && finish <= len))
    scheme_signal_error("substring: second index %d out of range [%d, %d]",
			finish, 0, len);

  str = scheme_alloc_string (finish-start, 0);
  for (i = 0; i < finish-start; i++)
    SCHEME_STR_VAL(str)[i] = chars[i+start];

  return (str);
}

static Scheme_Object *
string_append (int argc, Scheme_Object *argv[])
{
  Scheme_Object *naya;
  int i;

  naya = scheme_alloc_string (0, 0);
  for ( i=0 ; i<argc ; ++i )
    naya = scheme_append_string(naya, argv[i]);

  return naya;
}

Scheme_Object *
scheme_append_string(Scheme_Object *str1, Scheme_Object *str2)
{
  int len1, len2, i;
  char *chars1, *chars2;
  Scheme_Object *naya;

  if (!SCHEME_STRINGP(str1))
    scheme_wrong_type("string-append", "string", -1, 0, &str1);
  if (!SCHEME_STRINGP(str2))
    scheme_wrong_type("string-append", "string", -1, 0, &str2);

  chars1 = SCHEME_STR_VAL (str1);
  chars2 = SCHEME_STR_VAL (str2);
  len1 = SCHEME_STRTAG_VAL (str1);
  len2 = SCHEME_STRTAG_VAL (str2);
  naya = scheme_alloc_string (len1+len2, 0);
  for ( i=0 ; i<len1 ; ++i )
    SCHEME_STR_VAL(naya)[i] = chars1[i];

  for ( i=0 ; i<len2 ; ++i )
    SCHEME_STR_VAL(naya)[i+len1] = chars2[i];

  SCHEME_STR_VAL(naya)[len1+len2] = '\0';

  return naya;
}


static Scheme_Object *
string_to_list (int argc, Scheme_Object *argv[])
{
  int len, i;
  char *chars;
  Scheme_Object *first, *last, *pair;

  if (argc != 1)
    scheme_wrong_count("string->list", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("string->list", "string", 0, argc, argv);

  chars = SCHEME_STR_VAL(argv[0]);
  len = SCHEME_STRTAG_VAL(argv[0]);
  first = last = scheme_null;
  for ( i=0 ; i<len ; ++i ) {
    pair = scheme_make_pair (scheme_make_char (chars[i]), scheme_null);
    if (SCHEME_NULLP(first))
      first = pair;
    else
      SCHEME_CDR (last) = pair;
    last = pair;
  }

  return first;
}

static Scheme_Object *
list_to_string (int argc, Scheme_Object *argv[])
{
  int len, i;
  Scheme_Object *list, *str, *ch;

  if (argc != 1)
    scheme_wrong_count("list->string", 1, 1, argc, argv);

  list = argv[0];
  len = scheme_list_length (list);
  str = scheme_alloc_string (len, 0);
  i = 0;
  while (SCHEME_PAIRP (list))
    {
      ch = SCHEME_CAR (list);

      if (!SCHEME_CHARP(ch))
	scheme_wrong_type("list->string", "proper character list", 0, 
			  argc, argv);

      SCHEME_STR_VAL(str)[i] = SCHEME_CHAR_VAL(ch);
      i++;
      list = SCHEME_CDR (list);
    }  

  if (!SCHEME_NULLP(list))
    scheme_wrong_type("list->string", "proper character list", 0, argc, argv);

  return (str);
}

static Scheme_Object *
string_copy (int argc, Scheme_Object *argv[])
{
  Scheme_Object *naya;

  if (argc != 1)
    scheme_wrong_count("string-copy", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("string-copy", "string", 0, argc, argv);

  naya = scheme_make_sized_string(SCHEME_STR_VAL(argv[0]), 
				  SCHEME_STRTAG_VAL(argv[0]), 1);
  return naya;
}

static Scheme_Object *
string_fill (int argc, Scheme_Object *argv[])
{
  int len, i;
  char *chars, ch;

  if (argc != 2)
    scheme_wrong_count("string-fill!", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("string-fill!", "string", 0, argc, argv);
  if (!SCHEME_CHARP(argv[1]))
    scheme_wrong_type("string-fill!", "character", 1, argc, argv);
  
  chars = SCHEME_STR_VAL (argv[0]);
  ch = SCHEME_CHAR_VAL (argv[1]);
  len = SCHEME_STRTAG_VAL (argv[0]);
  for ( i=0 ; i<len ; ++i )
    chars[i] = ch;

  return scheme_void;
}

static int
strcmp_ci (char *str1, char *str2)
{
  while (*str1 && *str2) {
    if (toupper(*str1) == toupper(*str2)) {
      str1++;
      str2++;
    } else if (toupper(*str1) < toupper(*str2))
      return -1;
    else
      return 1;
  }

  if (*str1)
    return 1;
  else if (*str2)
    return -1;
  else
    return 0;
}

static Scheme_Object *
version (int argc, Scheme_Object *argv[])
{
  if (argc)
    scheme_wrong_count("version", 0, 0, argc, argv);

  return scheme_make_string(scheme_version());
}

static Scheme_Object *
banner (int argc, Scheme_Object *argv[])
{
  if (argc)
    scheme_wrong_count("banner", 0, 0, argc, argv);
  
  return scheme_make_string(scheme_banner());
}

char *scheme_version(void)
{
  return VERSION;
}

char *scheme_banner(void)
{
  return "Welcome to MzScheme version " VERSION 
    ", Copyright (c) 1995 Matthew Flatt\n";
}

#ifdef GETENV_FUNCTION
static Scheme_Object *sch_getenv(int argc, Scheme_Object *argv[])
{
  char *s;

  if (argc != 1)
    scheme_wrong_count("getenv", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("getenv", "string", 0, argc, argv);

  s = getenv(SCHEME_STR_VAL(argv[0]));

  if (s)
    return scheme_make_string(s);
  else
    return scheme_false;
}
#endif
