/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"

struct Scheme_Struct_Proc
{
  Scheme_Type type;
  char *type_name;
  char *func_name;
  enum {SCHEME_CONSTR, SCHEME_PRED, SCHEME_GETTER, SCHEME_SETTER} proc_type;
  int slot_num;
};
typedef struct Scheme_Struct_Proc Scheme_Struct_Proc;

typedef struct {
  Scheme_Type type;
  Scheme_Object *name;
  Scheme_Object *fields;
  int num_fields;
  int count;
  /* rest are used only by let-struct */
  int startPos;
  int allocate;
  int extra_slots;
  Scheme_Object *body;
} Struct_Info;

/* globals */
Scheme_Object *scheme_define_struct_syntax;
Scheme_Object *scheme_arity_at_least_constr;

/* locals */

static Scheme_Object *define_struct_syntax(Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *define_struct_expand(Scheme_Object *form, Scheme_Env *env, int depth);
static Scheme_Object *let_struct_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *let_struct_expand(Scheme_Object *form, Scheme_Env *env, int depth);

static Scheme_Structure *scheme_make_instance (char *type_name, int num_fields);
static Scheme_Object *scheme_make_struct_proc (char *type_name, char *func_name, int proc_type, int field_num);
static Scheme_Object *scheme_make_constructor (char *type_name, char *struct_name, int num_fields);
static Scheme_Object *scheme_make_pred (char *type_name, char *struct_name);
static Scheme_Object *scheme_make_getter (char *type_name, char *struct_name, 
					  char *field_name, int field);
static Scheme_Object *scheme_make_setter (char *type_name, char *struct_name,  
					  char *field_name, int field);
static char *type_name (char *struct_name);
static char *constructor_name (char *struct_name);
static char *pred_name (char *struct_name);
static char *getter_name (char *struct_name, char *field_name);
static char *setter_name (char *struct_name, char *field_name);

static Scheme_Object *do_define_struct (Scheme_Object *form, Scheme_Env *env);
static Scheme_Object *let_struct_execute(Scheme_Object *form, Scheme_Env *env);

static Scheme_Object *write_struct_info(Scheme_Object *obj);
static Scheme_Object *read_struct_info(Scheme_Object *obj);

static Scheme_Object **make_struct_values(Struct_Info *info, 
					  Scheme_Object **constructor);
static Scheme_Object **make_struct_names(Struct_Info *info);
static Scheme_Object *make_structure(Scheme_Object *struct_symbol,
				     Scheme_Object *field_symbols);

void
scheme_init_struct (Scheme_Env *env)
{
  int i;
  static int all_count;
  static Scheme_Object **all_values, **all_names;

  if (scheme_starting_up) {
    Struct_Info *info;

    scheme_register_syntax_executer("-ds", do_define_struct);
    scheme_register_syntax_executer("-ls", let_struct_execute);

    scheme_install_type_writer(scheme_struct_info_type, write_struct_info);
    scheme_install_type_reader(scheme_struct_info_type, read_struct_info);

    scheme_define_struct_syntax 
      = scheme_make_compiled_syntax(define_struct_syntax,define_struct_expand);

    info = (Struct_Info *)
      make_structure(scheme_intern_symbol("arity-at-least"),
		     scheme_make_pair(scheme_intern_symbol("value"),
				      scheme_null));
    
    all_count = info->count;
    all_names = make_struct_names(info);
    all_values = make_struct_values(info, &scheme_arity_at_least_constr);
  }

  scheme_add_global_keyword ("define-struct", scheme_define_struct_syntax, env);
  scheme_add_global_keyword ("let-struct", scheme_make_compiled_syntax(let_struct_syntax, let_struct_expand), env);  

  /* Add arity structure */
  for (i = 0; i < all_count; i++)
    scheme_add_global_constant(SCHEME_SYM_VAL(all_names[i]), all_values[i], 
			       env);
}

static void wrong_struct_type(const char *name, const char *expected,
			      const char *received,
			      int which, int argc,
			      Scheme_Object **argv)
{
  if (!strcmp(expected, received))
    scheme_signal_error("%s: expects args of type <%s>; "
			"given instance of a different <%s>",
			name, expected, received);
  else
    scheme_wrong_type(name, expected, which, argc, argv);
}

Scheme_Object *
scheme_apply_struct_proc (Scheme_Object *sp, int argc, Scheme_Object **args)
{
  Scheme_Struct_Proc *proc;

  proc = (Scheme_Struct_Proc *)sp;
  switch (proc->proc_type)
    {
    case SCHEME_CONSTR:
      {
	Scheme_Structure *inst;
	int i;

	inst = scheme_make_instance (proc->type_name, proc->slot_num);
	if (argc > proc->slot_num || argc < proc->slot_num)
	  scheme_wrong_count(proc->func_name, proc->slot_num,
			     proc->slot_num, argc, args);

	for (i = 0; i < argc; i++)
	  inst->slots[i] = args[i];

	return (Scheme_Object *)inst;
      }
    case SCHEME_PRED:
      if (argc != 1)
	scheme_wrong_count(proc->func_name, 1, 1, argc, args);
      if (SAME_TYPE(SCHEME_TYPE (args[0]), scheme_structure_type)
	  && (proc->type_name == ((Scheme_Structure *)args[0])->type_name))
	return scheme_true;
      else
	return scheme_false;
    case SCHEME_GETTER:
      {
	Scheme_Structure *inst;

	if (argc != 1)
	  scheme_wrong_count(proc->func_name, 1, 1, argc, args);

	if (NOT_SAME_TYPE(SCHEME_TYPE(args[0]), scheme_structure_type))
	  scheme_wrong_type(proc->func_name, proc->type_name, 0, argc, args);
	
	inst = (Scheme_Structure *)args[0];
	if (inst->type_name != proc->type_name)
	  wrong_struct_type(proc->func_name, proc->type_name, inst->type_name, 
			    0, argc, args);

	return inst->slots[proc->slot_num];
      }
    case SCHEME_SETTER:
      {
	Scheme_Structure *inst;

	if (argc != 2)
	  scheme_wrong_count(proc->func_name, 1, 1, argc, args);

	if (NOT_SAME_TYPE(SCHEME_TYPE(args[0]), scheme_structure_type))
	  scheme_wrong_type(proc->func_name, proc->type_name, 0, argc, args);
	
	inst = (Scheme_Structure *)args[0];
	if (inst->type_name != proc->type_name)
	  wrong_struct_type(proc->func_name, proc->type_name, inst->type_name,
			    0, argc, args);
	
	inst->slots[proc->slot_num] = args[1];

	return scheme_void;
      }
    default:
      scheme_signal_error("internal error: "
			  "unknown procedure for structure type %s",
			  proc->type_name);
      return scheme_void;
    }
}

void scheme_struct_get_arity(Scheme_Object *obj, short *mina, short *maxa)
{
  Scheme_Struct_Proc *proc;

  proc = (Scheme_Struct_Proc *)obj;
  switch (proc->proc_type) {
  case SCHEME_CONSTR:
    *mina = *maxa = proc->slot_num;
    break;
  case SCHEME_PRED:
    *mina = *maxa = 1;
    break;
  case SCHEME_GETTER:
  case SCHEME_SETTER:
    *mina = *maxa = 2;
    break;
  }
}

int scheme_equal_structs (Scheme_Object *obj1, Scheme_Object *obj2)
{
  Scheme_Structure *s1, *s2;
  int i;

  if (!SAME_TYPE(SCHEME_TYPE(obj1), scheme_structure_type)
      || !SAME_TYPE(SCHEME_TYPE(obj2), scheme_structure_type))
    return 0;

  s1 = (Scheme_Structure *)obj1;
  s2 = (Scheme_Structure *)obj2;

  if (s1->type_name != s2->type_name)
    return 0;

  for (i = s1->num_slots; i--; )
    if (!scheme_equal(s1->slots[i], s2->slots[i]))
      return 0;

  return 1;
}

#define NUM_BASE_VALUES 2
#define NUM_VALUES_PER_FIELD 2

static Scheme_Object **make_struct_values(Struct_Info *info, 
					  Scheme_Object **constructor)
{
  Scheme_Object **values, *field_symbols;
  char *struct_type_name, *struct_name, *field_name;
  int slot_num, pos;

  values = (Scheme_Object **)scheme_malloc_stubborn(info->count * sizeof(Scheme_Object *));

  struct_type_name = type_name(struct_name = SCHEME_SYM_VAL(info->name));
  
  values[0] = scheme_make_constructor(struct_type_name, struct_name, 
				      info->num_fields);
  if (constructor)
    *constructor = values[0];
  values[1] = scheme_make_pred(struct_type_name, struct_name);

  field_symbols = info->fields;

  pos = 2;
  slot_num = 0;
  while (pos < info->count)
    {
      field_name = SCHEME_SYM_VAL(SCHEME_CAR (field_symbols));
      values[pos] = scheme_make_getter (struct_type_name, struct_name, 
					field_name, slot_num);
      pos++;
      values[pos] = scheme_make_setter(struct_type_name, struct_name, 
				       field_name, slot_num);
      pos++;
      slot_num++;
      field_symbols = SCHEME_CDR(field_symbols);
    }

  scheme_end_stubborn_change((void *)values);
 
  return values;
}

static Scheme_Object **make_struct_names(Struct_Info *info)
{
  Scheme_Object **names, *field_symbols;
  char *struct_name, *field_name;
  int slot_num, pos;

  names = (Scheme_Object **)scheme_malloc_stubborn(info->count * sizeof(Scheme_Object *));

  struct_name = SCHEME_SYM_VAL(info->name);
  
  names[0] = scheme_intern_symbol(constructor_name(struct_name));
  names[1] = scheme_intern_symbol(pred_name(struct_name));

  field_symbols = info->fields;

  pos = 2;
  slot_num = 0;
  while (!SCHEME_NULLP(field_symbols))
    {
      field_name = SCHEME_SYM_VAL(SCHEME_CAR (field_symbols));
      names[pos] = scheme_intern_symbol(getter_name(struct_name, field_name));
      pos++;
      names[pos] = scheme_intern_symbol(setter_name(struct_name, field_name));
      pos++;
      field_symbols = SCHEME_CDR (field_symbols);
      slot_num++;
    }

  scheme_end_stubborn_change((void *)names);

  return names;
}

static Scheme_Object *make_structure(Scheme_Object *struct_symbol,
				     Scheme_Object *field_symbols)
{
  Struct_Info *info;
  int count;

  SCHEME_ASSERT(SCHEME_SYMBOLP(struct_symbol),
		"struct: name must be a symbol");

  info = (Struct_Info *)scheme_malloc(sizeof(Struct_Info));
  info->type = scheme_struct_info_type;

  info->name = struct_symbol;
  
  info->fields = field_symbols;

  count = 0;
  while (! SCHEME_NULLP(field_symbols))
    {
      count++;
      SCHEME_ASSERT (SCHEME_PAIRP (field_symbols), 
		     "structure: badly formed field list");
      field_symbols = SCHEME_CDR (field_symbols);
    }

  info->num_fields = count;
  info->count = (count * NUM_VALUES_PER_FIELD) + NUM_BASE_VALUES;

  info->startPos = info->allocate = info->extra_slots = 0;
  info->body = scheme_null;

  return (Scheme_Object *)info;
}

static Scheme_Object *
do_define_struct (Scheme_Object *form, Scheme_Env *env)
{
  Struct_Info *info;
  Scheme_Object **names, **values;
  int i;

  info = (Struct_Info *)form;

  names = make_struct_names(info);
  values = make_struct_values(info, NULL);

  for (i = 0; i < info->count; i++)
    scheme_add_global(SCHEME_SYM_VAL(names[i]), values[i], env);

  return scheme_void;
}

static Scheme_Object *
define_struct_syntax (Scheme_Object *form, Scheme_Env *env, 
		      Scheme_Compile_Info *rec)
{
  Scheme_Object *struct_symbol, *field_symbols;

  if (env->basic.next)
    scheme_signal_error("define-struct: only allowed at top-level");

  form = SCHEME_CDR(form);
  SCHEME_ASSERT(SCHEME_PAIRP(form), "define-struct: bad syntax");
  struct_symbol = SCHEME_CAR (form);
  form = SCHEME_CDR(form);
  SCHEME_ASSERT(SCHEME_PAIRP(form), "define-struct: bad syntax");
  field_symbols = SCHEME_CAR (form);
  form = SCHEME_CDR(form);
  SCHEME_ASSERT(SCHEME_NULLP(form), "define-struct: bad syntax");

  rec->max_let_depth = 0;
  rec->can_return_closure = 0;

  return scheme_make_syntax_compilation(do_define_struct,
					make_structure(struct_symbol, 
						       field_symbols));
}

static Scheme_Object *define_struct_expand(Scheme_Object *form, 
					   Scheme_Env *env, int depth)
{
  return form;
}

static Scheme_Object *
let_struct_execute (Scheme_Object *form, Scheme_Env *env)
{
  Struct_Info *info;
  Scheme_Object **values;
  Scheme_Env *frame;
  int i;

  info = (Struct_Info *)form;

  if (info->allocate) {
    frame = scheme_new_frame(info->count + info->extra_slots);
    env = scheme_extend_env(frame, env);
  } 

  values = make_struct_values(info, NULL);

  for (i = info->count; i--; )
    scheme_add_good_binding(i + info->startPos, values[i], env);

  return _scheme_tail_eval(info->body, env);
}

#define cons scheme_make_pair

static Scheme_Object *
do_let_struct_syntax (Scheme_Object *forms, Scheme_Env *env, 
		      Scheme_Compile_Info *rec, int depth)
{
  Scheme_Object *struct_symbol, *field_symbols, *form;
  Scheme_Object *obj, **names;
  Scheme_Env *frame, *orig_env;
  int i;
  Struct_Info *info;

  form = SCHEME_CDR(forms);
  SCHEME_ASSERT(SCHEME_PAIRP(form), "let-struct: bad syntax");
  struct_symbol = SCHEME_CAR (form);
  form = SCHEME_CDR(form);
  SCHEME_ASSERT(SCHEME_PAIRP(form), "let-struct: bad syntax");
  field_symbols = SCHEME_CAR (form);
  form = SCHEME_CDR(form);

  obj = make_structure(struct_symbol, field_symbols);

  info = (Struct_Info *)obj;

  if (rec) {
    if (rec->base_let_depth < 0) {
      info->allocate = 1;
      info->startPos = 0;
      rec->base_let_depth = info->count;
    } else {
      info->startPos = rec->base_let_depth;
      info->allocate = 0;
      rec->base_let_depth += info->count;
    }
  } else {
    info->allocate = 1;
    info->startPos = 0;
  }

  if (info->allocate) {
    orig_env = NULL;
    frame = scheme_new_compilation_frame(info->count + info->startPos,
					 (depth < 0) ? SCHEME_ANNOTATING : 0);
    env = scheme_extend_env(frame, env);
  } else {
    orig_env = env;
    env = frame = scheme_lengthen_compilation_frame(info->count, env);
  }

  names = make_struct_names(info);

  for (i = 0; i < info->count; i++)
    scheme_add_compilation_binding(i + info->startPos, names[i], frame);

  if (rec) {
    info->body = scheme_compile_sequence(form, env, rec);

    info->extra_slots = rec->max_let_depth;
    
    rec->max_let_depth += info->count;
    
   if (orig_env)
     scheme_frame_backinfo(orig_env, env);

    return scheme_make_syntax_compilation(let_struct_execute, obj);
  } else
    return cons(SCHEME_CAR(forms),
		cons(struct_symbol,
		     cons(field_symbols,
			  scheme_expand_list(form, env, depth))));
}

static Scheme_Object *
let_struct_syntax (Scheme_Object *form, Scheme_Env *env, 
		   Scheme_Compile_Info *rec)
{
  return do_let_struct_syntax(form, env, rec, 0);
}

static Scheme_Object *
let_struct_expand(Scheme_Object *form, Scheme_Env *env, int depth)
{
  return do_let_struct_syntax(form, env, NULL, depth);
}

static Scheme_Structure *
scheme_make_instance (char *type_name, int num_fields)
{
  Scheme_Structure *inst;

  inst = (Scheme_Structure *)scheme_malloc_stubborn(sizeof(Scheme_Structure));

  inst->type = scheme_structure_type;
  inst->type_name = type_name;
  inst->num_slots = num_fields;
  inst->slots = (Scheme_Object **) scheme_malloc (num_fields * sizeof (Scheme_Object*));

  scheme_end_stubborn_change((void *)inst);

  return (inst);
}

static Scheme_Object *
scheme_make_struct_proc(char *type_name, char *func_name, 
			int proc_type, int field_num)
{
  Scheme_Struct_Proc *proc;

  proc = 
    (Scheme_Struct_Proc *)scheme_malloc_stubborn(sizeof (Scheme_Struct_Proc));
  proc->type = scheme_struct_proc_type;
  proc->type_name = type_name;
  proc->func_name = func_name;
  proc->proc_type = proc_type;
  proc->slot_num = field_num;
  scheme_end_stubborn_change((void *)proc);

  return (Scheme_Object *)proc;
}

static Scheme_Object *
scheme_make_constructor(char *type_name, char *struct_name, int num_fields)
{
  return (scheme_make_struct_proc (type_name, constructor_name(struct_name), 
				   SCHEME_CONSTR, num_fields));
}

static Scheme_Object *
scheme_make_pred (char *type_name, char *struct_name)
{
  return scheme_make_struct_proc(type_name, pred_name(struct_name), 
				 SCHEME_PRED, 0);
}

static Scheme_Object *
scheme_make_getter (char *type_name, char *struct_name, char *field_name, 
		    int field)
{
  return (scheme_make_struct_proc (type_name, 
				   getter_name(struct_name, field_name), 
				   SCHEME_GETTER, field));
}

static Scheme_Object *
scheme_make_setter (char *type_name, char *struct_name, char *field_name, 
		    int field)
{
  return (scheme_make_struct_proc (type_name, 
				   setter_name(struct_name, field_name), 
				   SCHEME_SETTER, field));
}

static char *
type_name (char *struct_name)
{
  int orig_len;
  char *name;

  orig_len = strlen (struct_name);
#define  add_len 10	/* strlen("structure:") */
  name = (char *)scheme_malloc_atomic(sizeof(char)*(orig_len + add_len + 1));
  strcpy(name, "structure:");
  strcpy(name + add_len, struct_name);

  return (name);
}

static char *
constructor_name (char *struct_name)
{
  int orig_len, make_len;
  char *name;

  orig_len = strlen (struct_name);
  make_len = 5;			/* strlen ("make-") */
  name = (char *)scheme_malloc_atomic(sizeof(char)*(orig_len + make_len + 1));
  strcpy (name, "make-");
  strcat (name, struct_name);
  return (name);
}

static char *
pred_name (char *struct_name)
{
  int orig_len;
  char *name;

  orig_len = strlen (struct_name);
  name = (char *)scheme_malloc_atomic(sizeof(char)*(orig_len + 1 + 1));
  strcpy (name, struct_name);
  name[orig_len] = '?';
  name[orig_len+1] = '\0';
  return (name);
}

static char *
getter_name (char *struct_name, char *field_name)
{
  int name_len, field_len, dash_len;
  char *name;

  name_len = strlen (struct_name);
  field_len = strlen (field_name);
  dash_len = 1;			/* strlen ("-") */
  name = (char *)scheme_malloc_atomic(sizeof(char)*(name_len
						    + dash_len 
						    + field_len + 1));
  strcpy (name, struct_name);
  strcat (name, "-");
  strcat (name, field_name);
  return (name);
}

static char *
setter_name (char *struct_name, char *field_name)
{
  int set_len, name_len, field_len, dash_len, bang_len;
  char *name;

  name_len = strlen (struct_name);
  field_len = strlen (field_name);
  set_len = 4;			/* strlen ("set-") */
  dash_len = 1;			/* strlen ("-") */
  bang_len = 1;			/* strlen ("!") */
  name = (char *) 
    scheme_malloc_atomic(sizeof(char)*(set_len+name_len+dash_len+field_len
				       +bang_len + 1));
  strcpy (name, "set-");
  strcat (name, struct_name);
  strcat (name, "-");
  strcat (name, field_name);
  strcat (name, "!");
  return (name);
}

/************************************************************************/

#define cons scheme_make_pair

static Scheme_Object *write_struct_info(Scheme_Object *obj)
{
  Struct_Info *info;

  info = (Struct_Info *)obj;

  return cons(scheme_make_integer(info->count),
	      cons(scheme_make_integer(info->num_fields),
		   cons(scheme_make_integer(info->startPos),
			cons(scheme_make_integer(info->extra_slots),
			     cons(info->allocate ? scheme_true : scheme_false,
				  cons(info->body,
				       cons(info->name, info->fields)))))));
}

static Scheme_Object *read_struct_info(Scheme_Object *obj)
{
  Scheme_Object *v;
  Struct_Info *info;

#define BAD_CS "bad compiled structure info"

  info = (Struct_Info *)scheme_malloc_stubborn(sizeof(Struct_Info));
  info->type = scheme_struct_info_type;

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CS);
  v = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  SCHEME_ASSERT(SCHEME_INTP(v), BAD_CS);
  info->count = SCHEME_INT_VAL(v);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CS);
  v = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  SCHEME_ASSERT(SCHEME_INTP(v), BAD_CS);
  info->num_fields = SCHEME_INT_VAL(v);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CS);
  v = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  SCHEME_ASSERT(SCHEME_INTP(v), BAD_CS);
  info->startPos = SCHEME_INT_VAL(v);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CS);
  v = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  SCHEME_ASSERT(SCHEME_INTP(v), BAD_CS);
  info->extra_slots = SCHEME_INT_VAL(v);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CS);
  v = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  SCHEME_ASSERT(SCHEME_BOOLP(v), BAD_CS);
  info->allocate = SAME_OBJ(v, scheme_true);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CS);
  info->body = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  
  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CS);
  info->name = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  
  SCHEME_ASSERT(SCHEME_LISTP(obj), BAD_CS);
  info->fields = obj;
  
  scheme_end_stubborn_change((void *)info);

  return (Scheme_Object *)info;
}

