/*
 * File:        wx_mpbd.cc
 * Purpose:     wxMediaPasteboard implementation
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */

#include "wx.h"
#include "wx_media.h"
#include "wx_ptreq.h"

#define LINE_PP 100

#define DOT_WIDTH 5
#define HALF_DOT_WIDTH 2

#ifndef ROUND
#include <math.h>
#define ROUND(x) (int)floor(x)
#endif

#define SnipLoc(snip) ((wxSnipLocation *)snipLocationList->Find((long)snip)->Data())

inline Bool Inbox(float lx, float x)
{ 
  return ((lx - HALF_DOT_WIDTH <= x)
	  && (lx - HALF_DOT_WIDTH + DOT_WIDTH >= x));
}

class wxSnipLocation : public wxObject
{
 public:
  float x, y, w, h, r, b, hm, vm;
  Bool selected, needResize;
  wxSnip *snip;

  void Resize(wxDC *dc);
};

wxBrush *blackBrush = NULL, *rbBrush = NULL;
wxPen *invisiPen = NULL, *rbPen = NULL;

/**********************************************************************/

wxMediaPasteboard::wxMediaPasteboard()
{
  sizeCacheInvalid = TRUE;
  updateNonemtpy = FALSE;
  locked = FALSE;

  snips = lastSnip = NULL;
  snipLocationList = new wxList(wxKEY_INTEGER);
  snipLocationList->DeleteContents(TRUE);

  sequence = 0;

  __type = wxTYPE_MEDIA_PASTEBOARD;
  bufferType = wxPASTEBOARD_BUFFER;

  totalWidth = totalHeight = realWidth = realHeight = 0;

  dragable = TRUE;

  sequenceStreak = FALSE;

  dragging = rubberband = FALSE;

  if (!blackBrush) {
    blackBrush = wxTheBrushList->FindOrCreateBrush("black", wxSOLID);
    invisiPen = wxThePenList->FindOrCreatePen("black", 1, wxTRANSPARENT);
    rbBrush = wxTheBrushList->FindOrCreateBrush("black", wxTRANSPARENT);
    rbPen = wxThePenList->FindOrCreatePen("black", 1, wxDOT);
  }

  bitmap = NULL;
  offscreen = new wxMemoryDC();
  bmHeight = bmWidth = 0;

  snipAdmin = new wxStandardSnipAdmin(this);

  needResize = FALSE;

  keepSize = FALSE;
}

wxMediaPasteboard::~wxMediaPasteboard()
{
  wxSnip *snip, *next;
  for (snip = snips; snip; snip = next) {
    next = snip->next;
    delete snip;
  }

  delete snipLocationList;

  delete offscreen;

  delete snipAdmin;
}

void wxMediaPasteboard::RubberBand(float x, float y, float w, float h)
{
  wxPen *oldPen;
  wxBrush *oldBrush;
  wxDC *dc;
  float vx, vy, vw, vh, b, r, dx, dy;

  if (!admin)
    return;

  if (!w & !h)
    return;

  if (w < 0) {
    x += w;
    w = -w;
  }
  r = x + w;
  if (h < 0) {
    y += h;
    h = -h;
  }
  b = y + h;

  admin->GetView(&vx, &vy, &vw, &vh);

  if (x < vx)
    x = vx;
  if (y < vy)
    y = vy;
  if (r > vx + vw)
    r = vx + vw;
  if (b > vy + vh)
    b = vy + vh;

  if (x >= r || y >= b)
    return;

  dc = admin->GetDC(&dx, &dy);
  
  oldPen = dc->GetPen();
  oldBrush = dc->GetBrush();
  dc->SetPen(rbPen);
  dc->SetBrush(rbBrush);
  dc->SetLogicalFunction(wxXOR);
  
  dc->DrawRectangle(x - dx, y - dy, r - x, b - y);
  
  dc->SetLogicalFunction(wxCOPY);
  dc->SetPen(oldPen);
  dc->SetBrush(oldBrush);;
}


void wxMediaPasteboard::OnEvent(wxMouseEvent &event)
{
  float x, y, scrollx, scrolly;
  wxSnip *snip;
  wxSnipLocation *loc;
  wxDC *dc;

  if (event.ButtonDown() || caretSnip) {
    /* First, find clicked-on snip: */
    x = event.x;
    y = event.y;
    
    dc = admin->GetDC(&scrollx, &scrolly);
    y += scrolly;
    x += scrollx;
  } else
    dc = NULL;

  if (event.ButtonDown())
    snip = FindSnip(x, y);
  else
    snip = caretSnip;

  if (caretSnip && PTREQ(snip, caretSnip)) {
    loc = SnipLoc(caretSnip);
    caretSnip->OnEvent(dc, loc->x, loc->y, x - scrollx, y - scrolly, event);
    return;
  }

  OnLocalEvent(event);
}

void wxMediaPasteboard::OnDefaultEvent(wxMouseEvent& event)
{
  float x, y, scrollx, scrolly;
  wxSnip *snip;
  wxSnipLocation *loc;
  wxDC *dc;

  /* First, find clicked-on snip: */
  x = event.x;
  y = event.y;
  
  dc = admin->GetDC(&scrollx, &scrolly);
  y += scrolly;
  x += scrollx;

  if (event.ButtonDown()) {
    snip = FindSnip(x, y);

    resizing = NULL;
    if (dragable) {
      if (snip) {
	loc = SnipLoc(snip);
	origX = loc->x;
	origY = loc->y;
	origW = loc->w;
	origH = loc->h;
	if (!loc->selected) {
	  if (!event.shiftDown)
	    NoSelected();
	  AddSelected(snip);
	  dragging = TRUE;
	} else if (FindDot(loc, x, y, &sizedxm, &sizedym)) {
	  resizing = snip;
	  dragging = TRUE;
	} else if (snip->flags & wxSNIP_HANDLES_EVENTS) {
	  long interval;

	  interval = event.timeStamp - lastTime;
	  if (interval < 0)
	    interval = -interval;
	  if (interval < map->GetDoubleClickInterval()) {
	    NoSelected();
	    SetCaretOwner(snip);
	  }
	} else
	  dragging = TRUE;
	lastTime = event.timeStamp;
      } else {
	if (!event.shiftDown)
	  NoSelected();
	SetCaretOwner(NULL);
	rubberband = TRUE;
      }
      startX = lastX = x;
      startY = lastY = y;
    } else
      SetCaretOwner(snip);
  }

  if (dragable) {
    if (event.Dragging()) {
      if (rubberband) {
	/* Erase old */
	RubberBand(startX, startY, lastX - startX, lastY - startY);
	/* Draw new: */
	RubberBand(startX, startY, x - startX, y - startY);
      } else {
	keepSize = TRUE;
	if (resizing)
	  DoEventResize(x, y);
	if (!resizing)
	  Move(x - lastX, y - lastY);
      }
      lastX = x;
      lastY = y;
    } else if (event.ButtonUp()) {
      if (dragging) {
	if (resizing) {
	  BeginEditSequence();
	  /* Move & resize back without Undo */
	  if (sizedxm < 0 || sizedym < 0)
	    MoveTo(resizing, origX, origY);
	  Resize(resizing, origW, origH);
	  dragging = FALSE;
	  /* Re-move and re-size with undo: */
	  DoEventResize(lastX, lastY);
	  EndEditSequence();
	} else {
	  BeginEditSequence();
	  /* Move back without Undo */
	  Move(startX - lastX, startY - lastY);
	  dragging = FALSE;
	  /* Re-move with undo: */
	  Move(lastX - startX, lastY - startY);
	  EndEditSequence();
	}
      } else if (rubberband) {
	rubberband = FALSE;
	RubberBand(startX, startY, lastX - startX, lastY - startY);
	AddSelected(startX, startY, lastX - startX, lastY - startY);
      }

      keepSize = FALSE;
      UpdateAll();
    }
  }
}

void wxMediaPasteboard::OnChar(wxKeyEvent &event)
{
  float x, y, scrollx, scrolly;
  wxSnipLocation *loc;
  wxDC *dc;

  x = event.x;
  y = event.y;

  dc = admin->GetDC(&scrollx, &scrolly);
  y += scrolly;
  x += scrollx;

  if (caretSnip) {
    loc = SnipLoc(caretSnip);
    caretSnip->OnChar(dc, loc->x, loc->y, x - scrollx, y - scrolly, event);
    return;
  }

  OnLocalChar(event);
}

void wxMediaPasteboard::OnDefaultChar(wxKeyEvent &event)
{
  long code;

  code = event.KeyCode();

  switch(code) {
    case WXK_BACK:
    case WXK_DELETE:
      Delete();
      break;
    case WXK_RIGHT:
      Move(1, 0);
      break;
    case WXK_LEFT:
      Move(-1, 0);
      break;
    case WXK_UP:
      Move(0, -1);
      break;
    case WXK_DOWN:
      Move(0, 1);
      break;
    }
}

void wxMediaPasteboard::DoEventResize(float eventX, float eventY)
{
  float Dx, Dy, w, h, x, y;
  
  Dx = eventX - startX;
  Dy = eventY - startY;

  w = origW + Dx * sizedxm;
  h = origH + Dy * sizedym;
  
  x = origX;
  if (sizedxm < 0) {
    x += (eventX - startX);
    if (w < 0)
      x += w;
  }
  y = origY;
  if (sizedym < 0) {
    y += (eventY - startY);
    if (h < 0)
      y += h;
  }
  
  if (w < 0)
    w = 0;
  if (h < 0)
    h = 0;
  
  BeginEditSequence();
  
  if (!Resize(resizing, w, h))
    resizing = NULL;
  
  if (resizing && (sizedxm < 0 || sizedym < 0))
    MoveTo(resizing, x, y);
  
  EndEditSequence();
}

/***************************************************************************/

void wxMediaPasteboard::SetSelected(wxSnip *snip)
{
  BeginEditSequence();
  NoSelected();
  AddSelected(snip);
  EndEditSequence();
}

void wxMediaPasteboard::AddSelected(wxSnip *snip)
{
  wxNode *node;
  wxSnipLocation *loc;

  if (node = snipLocationList->Find((long)snip)) {
    loc = (wxSnipLocation *)node->Data();
    if (!loc->selected) {
      loc->selected = TRUE;
      UpdateLocation(loc);
    }
  }
}

void wxMediaPasteboard::AddSelected(float x, float y, float w, float h)
{
  wxNode *node;
  wxSnipLocation *loc;
  float r, b;

  if (w < 0) {
    x += w;
    w = -w;
  }
  if (h < 0) {
    y += h;
    h = -h;
  }

  r = x + w;
  b = y + h;

  BeginEditSequence();

  for (node = snipLocationList->First(); node; node = node->Next()) {
    loc = (wxSnipLocation *)node->Data();
    if (!loc->selected
	&& (loc->x <= r)
	&& (loc->y <= b)
	&& (loc->r >= x)
	&& (loc->b >= y)) {
      loc->selected = TRUE;
      UpdateLocation(loc);
    }
  }
  
  EndEditSequence();
}
				  
void wxMediaPasteboard::NoSelected()
{
  wxNode *node;
  wxSnipLocation *loc;

  BeginEditSequence();

  for (node = snipLocationList->First(); node; node = node->Next()) {
    loc = (wxSnipLocation *)node->Data();
    if (loc->selected) {
      loc->selected = FALSE;
      UpdateLocation(loc);
    }
  }
  
  EndEditSequence();
}

void wxMediaPasteboard::Insert(wxSnip *snip, wxSnip *before, float x, float y)
{
  wxSnipLocation *loc;
  wxSnip *search;

  if (snip->IsOwned())
    return;

  if (!snip->snipclass)
    wxMessageBox("Inserting a snip without a class."
		 " Data will be lost if you try to save the file.", 
		 "Warning");

  BeginEditSequence();
  if (!OnInsert(snip, before, x, y)) {
    EndEditSequence();
    return;
  }

  for (search = snips; search && (search != before); search = search->next);
  
  snip->next = search;
  if (snip->next) {
    snip->prev = search->prev;
    snip->next->prev = snip;
  } else {
    snip->prev = lastSnip;
    lastSnip = snip;
  }
  if (snip->prev)
    snip->prev->next = snip;
  else
    snips = snip;

  loc = new wxSnipLocation;
  loc->x = x;
  loc->y = y;
  loc->snip = snip;
  loc->needResize = TRUE;
  needResize = TRUE;
  loc->selected = FALSE;
  snipLocationList->Append((long)snip, loc);

  snip->style = styleList->Convert(snip->style);
  if (PTREQ(snip->style, styleList->BasicStyle()))
    snip->style = styleList->FindNamedStyle(STD_STYLE);

  AddUndo(new wxInsertSnipRecord(snip, sequenceStreak));
  if (sequence)
    sequenceStreak = TRUE;

  snip->SetAdmin(snipAdmin);

  snip->SizeCacheInvalid();

  AfterInsert(snip, before, x, y);
  changed = TRUE;

  UpdateLocation(loc);
  EndEditSequence();
}

void wxMediaPasteboard::Insert(wxSnip *snip, float x, float y)
{
  Insert(snip, snips, x, y);
}

void wxMediaPasteboard::Insert(wxSnip *snip)
{
  float x, y;

  GetCenter(&x, &y);
  Insert(snip, x, y);
}

void wxMediaPasteboard::Insert(wxSnip *snip, wxSnip *before)
{
  float x, y;

  GetCenter(&x, &y);
  Insert(snip, before, x, y);
}

void wxMediaPasteboard::Delete()
{
  wxNode *node;
  wxSnipLocation *loc;
  wxDeleteSnipRecord *del;

  del = new wxDeleteSnipRecord(sequenceStreak);
  if (sequence)
    sequenceStreak = TRUE;

  BeginEditSequence();
  for (node = snipLocationList->First(); node; node = node->Next()) {
    loc = (wxSnipLocation *)node->Data();
    if (loc->selected) 
      _Delete(loc->snip, del);
  }
  EndEditSequence();

  AddUndo(del);
}

void wxMediaPasteboard::Erase()
{
  wxSnip *snip, *next;
  wxDeleteSnipRecord *del;

  del = new wxDeleteSnipRecord(sequenceStreak);
  if (sequence)
    sequenceStreak = TRUE;

  BeginEditSequence();
  for (snip = snips; snip; snip = next) {
    next = snip->next;
    _Delete(snip, del);
  }
  EndEditSequence();

  AddUndo(del);
}

void wxMediaPasteboard::_Delete(wxSnip *del_snip,
				wxDeleteSnipRecord *del)
{
  wxSnip *snip;
  wxNode *node;
  wxSnipLocation *loc;

  for (snip = snips; snip; snip = snip->next) {
    if (PTREQ(snip, del_snip)) {
      BeginEditSequence();

      if (!OnDelete(del_snip)) {
	EndEditSequence();
	return;
      }

      UpdateSnip(del_snip);

      if (!snip->prev)
	snips = snip->next;
      else
	snip->prev->next = snip->next;
      if (!snip->next)
	lastSnip = snip->prev;
      else
	snip->next->prev = snip->prev;

      node = snipLocationList->Find((long)snip);
      snipLocationList->DeleteNode(node);
      loc = (wxSnipLocation *)node->Data();
      if (del)
	del->InsertSnip(snip, snip->next, loc->x, loc->y);
      snip->next = snip->prev = NULL;

      snip->SetAdmin(NULL);

      AfterDelete(del_snip);
      changed = TRUE;

      EndEditSequence();
    }
  }  

}

void wxMediaPasteboard::Delete(wxSnip *del_snip)
{
  wxDeleteSnipRecord *del;

  del = new wxDeleteSnipRecord(sequenceStreak);
  if (sequence)
    sequenceStreak = TRUE;

  _Delete(del_snip, del);

  AddUndo(del);
}

void wxMediaPasteboard::Remove(wxSnip *del_snip)
{
  _Delete(del_snip, NULL);
}

void wxMediaPasteboard::MoveTo(wxSnip *snip, float x, float y)
{
  wxNode *node;
  wxSnipLocation *loc;
  wxMoveSnipRecord *rec;

  if (node = snipLocationList->Find((long)snip)) {
    BeginEditSequence();
    if (!OnMoveTo(snip, x, y, dragging)) {
      EndEditSequence();
      return;
    }

    loc = (wxSnipLocation *)node->Data();
    UpdateLocation(loc);

    if (!dragging) {
      rec = new wxMoveSnipRecord(loc->snip, loc->x, loc->y, 
				 FALSE, sequenceStreak);
      if (sequence)
	sequenceStreak = TRUE;
      AddUndo(rec);
    }

    loc->x = x;
    loc->y = y;
    loc->r = x + loc->w;
    loc->b = y + loc->h;
    loc->hm = x + loc->w/2;
    loc->vm = y + loc->h/2;
    UpdateLocation(loc);

    AfterMoveTo(snip, x, y, dragging);

    needResize = TRUE;

    EndEditSequence();
  }
}

void wxMediaPasteboard::Move(wxSnip *snip, float dx, float dy)
{
  wxNode *node;
  wxSnipLocation *loc;

  if (node = snipLocationList->Find((long)snip)) {
    loc = (wxSnipLocation *)node->Data();
    MoveTo(snip, loc->x + dx, loc->y + dy);
  }
}

void wxMediaPasteboard::Move(float dx, float dy)
{
  wxNode *node;
  wxSnipLocation *loc;

  BeginEditSequence();

  for (node = snipLocationList->First(); node; node = node->Next()) {
    loc = (wxSnipLocation *)node->Data();
    if (loc->selected)
      Move(loc->snip, dx, dy);
  }

  EndEditSequence();
}

Bool wxMediaPasteboard::Resize(wxSnip *snip, float w, float h)
{
  wxNode *node;
  wxSnipLocation *loc;
  float oldw, oldh;
  Bool rv;
  
  if (!admin)
    return FALSE;

  if (!(node = snipLocationList->Find((long)snip)))
    return FALSE;

  loc = (wxSnipLocation *)node->Data();
  oldw = loc->w;
  oldh = loc->h;

  BeginEditSequence();
  if (!OnResize(snip, w, h)) {
    EndEditSequence();
    return FALSE;
  }

  if (!snip->Resize(w, h))
    rv = FALSE;
  else {
    if (!dragging) {
      AddUndo(new wxResizeSnipRecord(snip, oldw, oldh, sequenceStreak));
      if (sequence)
	sequenceStreak = TRUE;
    }
    rv = TRUE;
  }

  AfterResize(snip, w, h, rv);
  if (rv)
    changed = TRUE;
  EndEditSequence();

  return rv;
}

void wxMediaPasteboard::ChangeStyle(wxStyleDelta *delta)
{ 
  ChangeStyle(delta, NULL);
}

void wxMediaPasteboard::ChangeStyle(wxStyle *style, wxSnip *snip)
{
  _ChangeStyle(style, NULL, snip);
}

void wxMediaPasteboard::ChangeStyle(wxStyleDelta *delta, wxSnip *snip)
{
  _ChangeStyle(NULL, delta, snip);
}
 
void wxMediaPasteboard::_ChangeStyle(wxStyle *style, wxStyleDelta *delta, 
				     wxSnip *snip)
{
  wxNode *node;
  wxSnipLocation *loc;
  wxStyleChangeSnipRecord *rec;

  rec = new wxStyleChangeSnipRecord(sequenceStreak);
  if (sequence)
    sequenceStreak = TRUE;

  if (!style && !delta)
    style = styleList->FindNamedStyle(STD_STYLE);

  BeginEditSequence();

  if (snip) {
    rec->AddStyleChange(snip, snip->style);
    if (style)
      snip->style = style;
    else
      snip->style = styleList->FindOrCreateStyle(snip->style, delta);
    snip->SizeCacheInvalid();
    UpdateSnip(snip);
  } else {
    for (node = snipLocationList->First(); node; node = node->Next()) {
      loc = (wxSnipLocation *)node->Data();
      if (loc->selected) {
	rec->AddStyleChange(loc->snip, loc->snip->style);
	if (style)
	  loc->snip->style = style;
	else
	  loc->snip->style = styleList->FindOrCreateStyle(loc->snip->style, 
							  delta);
	loc->snip->SizeCacheInvalid();
	loc->needResize = TRUE;
	needResize = TRUE;
	UpdateLocation(loc);
      }
    }
  }
  
  EndEditSequence();

  AddUndo(rec);
}

void wxMediaPasteboard::Raise(wxSnip *snip)
{
}

void wxMediaPasteboard::Lower(wxSnip *snip)
{
}

void wxMediaPasteboard::SetBefore(wxSnip *snip, wxSnip *before)
{
}

void wxMediaPasteboard::SetAfter(wxSnip *snip, wxSnip *after)
{
}

/***************************************************************************/

Bool wxMediaPasteboard::FindDot(wxSnipLocation *loc, float x, float y,
				float *dxm, float *dym)
{
  if (Inbox(loc->x, x)) {
    *dxm = -1;
    if (Inbox(loc->y, y))
      *dym = -1;
    else if (Inbox(loc->vm, y))
      *dym = 0;
    else if (Inbox(loc->b, y))
      *dym = 1;
    else
      return FALSE;
  } else if (Inbox(loc->hm, x)) {
    *dxm = 0;
    if (Inbox(loc->y, y))
      *dym = -1;
    else if (Inbox(loc->b, y))
      *dym = 1;
    else
      return FALSE;
  } else if (Inbox(loc->r, x)) {
    *dxm = 1;
    if (Inbox(loc->y, y))
      *dym = -1;
    else if (Inbox(loc->vm, y))
      *dym = 0;
    else if (Inbox(loc->b, y))
      *dym = 1;
    else
      return FALSE;
  } else
    return FALSE;

  return TRUE;
}

wxSnip *wxMediaPasteboard::FindSnip(float x, float y)
{
  wxSnip *snip;
  wxSnipLocation *loc;
  float dym, dxm;

  for (snip = snips; snip; snip = snip->next) {
    loc = SnipLoc(snip);
    if (loc->x <= x && loc->y <= y
	&& loc->r >= x && loc->b >= y)
      return snip;
    else if (loc->selected && FindDot(loc, x, y, &dxm, &dym))
      return snip;
  }

  return NULL;
}

wxSnip *wxMediaPasteboard::FindFirstSnip(void)
{
  return snips;
}

Bool wxMediaPasteboard::IsSelected(wxSnip *asnip)
{
  wxSnip *snip;
  wxSnipLocation *loc;

  for (snip = snips; snip; snip = snip->next) {
    if (PTREQ(asnip, snip)) {
      loc = SnipLoc(snip);
      return loc->selected;
    }
  }

  return FALSE;
}

wxSnip *wxMediaPasteboard::FindNextSelectedSnip(wxSnip *start)
{
  wxSnip *snip;
  wxSnipLocation *loc;

  if (!start)
    snip = snips;
  else
    snip = start;

  for (snip = snips; snip; snip = snip->next) {
    loc = SnipLoc(snip);
    if (loc->selected)
      return snip;
  }

  return NULL;
}

/***************************************************************************/

void wxMediaPasteboard::Draw(wxDC *dc, float dx, float dy, 
			     float cx, float cy, float cw, float ch, 
			     Bool show_caret)
{
  wxSnip *snip;
  wxStyle *oldstyle = NULL;
  wxSnipLocation *loc;
  float cr, cb, x, y, r, b, hm, vm, dcx, dcy, dcr, dcb;

  if (!admin)
    return;

  locked = TRUE;

  dcx = cx + dx;
  dcy = cy + dy;

  cr = cx + cw;
  cb = cy + ch;

  dcr = dcx + cw;
  dcb = dcy + ch;

  dc->SetLogicalFunction(wxCLEAR);
  dc->DrawRectangle(dcx, dcy, cw, ch);
  dc->SetLogicalFunction(wxCOPY);

  OnPaint(TRUE, dc, cx, cy, cr, cb, dx, dy, show_caret);

  for (snip = lastSnip; snip; snip = snip->prev) {
    loc = SnipLoc(snip);

    if ((loc->x <= cr)
	&& (loc->y <= cb)
	&& (loc->r >= cx)
	&& (loc->b >= cy)) {
      snip->style->SwitchTo(dc, oldstyle);
      oldstyle = snip->style;

      x = loc->x + dx;
      y = loc->y + dy;
      
      snip->Draw(dc, x, y, dcx, dcy, dcr, dcb, dx, dy, 
		 PTREQ(snip, caretSnip) && show_caret);

      if (ownCaret && loc->selected) {
	wxBrush *oldbrush;
	wxPen *oldpen;

	oldbrush = dc->GetBrush();
	oldpen = dc->GetPen();
	dc->SetBrush(blackBrush);
	dc->SetPen(invisiPen);
	dc->SetLogicalFunction(wxXOR);

	r = loc->r + dx;
	b = loc->b + dy;
	hm = loc->hm + dx;
	vm = loc->vm + dy;

	dc->DrawRectangle(x - HALF_DOT_WIDTH, 
			  y - HALF_DOT_WIDTH, 
			  DOT_WIDTH, DOT_WIDTH);
	dc->DrawRectangle(hm - HALF_DOT_WIDTH, 
			  y - HALF_DOT_WIDTH, 
			  DOT_WIDTH, DOT_WIDTH);
	dc->DrawRectangle(r - HALF_DOT_WIDTH, 
			  y - HALF_DOT_WIDTH, 
			  DOT_WIDTH, DOT_WIDTH);
	dc->DrawRectangle(r - HALF_DOT_WIDTH, 
			  vm - HALF_DOT_WIDTH, 
			  DOT_WIDTH, DOT_WIDTH);
	dc->DrawRectangle(r - HALF_DOT_WIDTH, 
			  b - HALF_DOT_WIDTH, 
			  DOT_WIDTH, DOT_WIDTH);
	dc->DrawRectangle(hm - HALF_DOT_WIDTH, 
			  b - HALF_DOT_WIDTH, 
			  DOT_WIDTH, DOT_WIDTH);
	dc->DrawRectangle(x - HALF_DOT_WIDTH, 
			  b - HALF_DOT_WIDTH, 
			  DOT_WIDTH, DOT_WIDTH);
	dc->DrawRectangle(x - HALF_DOT_WIDTH, 
			  vm - HALF_DOT_WIDTH, 
			  DOT_WIDTH, DOT_WIDTH);

	dc->SetLogicalFunction(wxCOPY);
	dc->SetPen(oldpen);
	dc->SetBrush(oldbrush);
      }
    }
  }

  styleList->BasicStyle()->SwitchTo(dc, oldstyle);

  OnPaint(FALSE, dc, cx, cy, cr, cb, dx, dy, show_caret);

  locked = FALSE;
}

void wxMediaPasteboard::Refresh(float localx, float localy, float w, float h, 
				Bool show_caret)
{
  float dx, dy, ddx, ddy;
  wxDC *dc;

  if (!admin || locked)
    return;

  if (!h || !w)
    return;

  if (h > bmHeight || w > bmWidth) {
    wxBitmap *oldbm = bitmap;

    bmWidth = ROUND(w);
    bmHeight = ROUND(h);

    bitmap = new wxBitmap(bmWidth, bmHeight);

    offscreen->SelectObject(NULL);
    if (oldbm)
      delete oldbm;
  
    if (bitmap->Ok() && offscreen->Ok())
      offscreen->SelectObject(bitmap);
  } 

  dc = admin->GetDC(&dx, &dy);

  if (bitmap && bitmap->Ok() && offscreen->Ok()) {
    /* Need to make sure that difference between coordinates is
       integral; otherwise, roundoff error could affect drawing */
    ddx = (localx - dx) - (long)(localx - dx);
    if (ddx < 0)
      ddx = 1 + ddx;
    localx -= ddx;
    w += ddx;
    ddy = (localy - dy) - (long)(localy - dy);
    if (ddy < 0)
      ddy = 1 + ddy;
    localy -= ddy;
    h += ddy;

    Draw(offscreen, -localx, -localy, localx, localy, w, h, show_caret);
    dc->Blit(localx - dx, localy - dy, w, h, offscreen, 0, 0, wxCOPY);
  } else
    Draw(dc, -dx, -dy, localx, localy, w, h, show_caret);
}

void wxMediaPasteboard::CheckRecalc()
{
  float r, b;
  wxDC *dc;
  wxNode *node;
  wxSnipLocation *loc;
  
  if (!admin)
    return;

  dc = admin->GetDC();

  if (needResize) {
    /* Find right & bottom */
    r = b = 0;
    for (node = snipLocationList->First(); node; node = node->Next()) {
      loc = (wxSnipLocation *)node->Data();
      if (loc->needResize)
	loc->Resize(dc);
      if (loc->r + DOT_WIDTH > r)
	r = loc->r + DOT_WIDTH;
      if (loc->b + DOT_WIDTH > b)
	b = loc->b + DOT_WIDTH;
    }
  
    realWidth = r;
    realHeight = b;

    needResize = FALSE;
  }

  if (!keepSize) {
    if (realWidth != totalWidth || realHeight != totalHeight) {
      totalWidth = realWidth;
      totalHeight = realHeight;
      admin->Resized(FALSE);
    }
  }

  
}

void wxMediaPasteboard::Update(float x, float y, float w, float h)
{
  float r, b;

  r = x + w;
  b = y + h;

  if (!updateNonemtpy) {
    updateTop = y;
    updateLeft = x;
    updateBottom = b;
    updateRight = r;
    updateNonemtpy = TRUE;
  } else {
    if (y < updateTop)
      updateTop = y;
    if (x < updateLeft)
      updateLeft = x;
    if (updateBottom < 0) {
      if (b < 0 && b < updateBottom)
	updateBottom = b;
      else if (b > 0 && (-b) < updateBottom)
	updateBottom = -b;
    } else if (b < 0 || b > updateBottom)
      updateBottom = b;
    if (updateRight < 0) {
      if (r < 0 && r < updateRight)
	updateRight = r;
      else if (r > 0 && (-r) < updateRight)
	updateRight = -r;
    } else if (r < 0 || r > updateBottom)
      updateRight = r;
  }

  if (sequence || !admin)
    return;

  CheckRecalc();

  if (updateBottom < 0) {
    updateBottom = -updateBottom;
    if (updateBottom < realHeight)
      updateBottom = realHeight;
  }
  if (updateRight < 0) {
    updateRight = -updateRight;
    if (updateRight < realWidth)
      updateRight = realWidth;
  }

  if (updateTop != updateBottom || updateLeft != updateRight)
    admin->NeedsUpdate(updateLeft, updateTop, 
		       updateRight - updateLeft, updateBottom - updateTop);

  updateNonemtpy = FALSE;

  if (changed) {
    changed = FALSE;
    OnChange();
  }
}

void wxMediaPasteboard::UpdateLocation(wxSnipLocation *loc)
{
  if (admin) {
    if (loc->needResize)
      loc->Resize(admin->GetDC());
    Update(loc->x - HALF_DOT_WIDTH, loc->y - HALF_DOT_WIDTH, 
	   loc->w + DOT_WIDTH, loc->h + DOT_WIDTH);
  }
}

void wxMediaPasteboard::UpdateSnip(wxSnip *snip)
{
  wxNode *node;
  wxSnipLocation *loc;

  if (node = snipLocationList->Find((long)snip)) {
    loc = (wxSnipLocation *)node->Data();
    UpdateLocation(loc);
  }
}

void wxMediaPasteboard::UpdateSelected()
{
  wxNode *node;
  wxSnipLocation *loc;

  BeginEditSequence();

  for (node = snipLocationList->First(); node; node = node->Next()) {
    loc = (wxSnipLocation *)node->Data();
    if (loc->selected)
      UpdateLocation(loc);
  }
  
  EndEditSequence();
}

void wxMediaPasteboard::UpdateAll()
{
  Update(0, 0, -1, -1);
}

void wxMediaPasteboard::UpdateNeeded()
{
  if (updateNonemtpy)
    Update(updateLeft, updateTop, 0, 0);
}

void wxMediaPasteboard::InvalidateBitmapCache(float x, float y, float w, float h)
{
  Update(x, y, w, h);
}

/***************************************************************************/

void wxMediaPasteboard::OwnCaret(Bool ownit)
{
  if (DoOwnCaret(ownit))
    UpdateSelected();
}

void wxMediaPasteboard::SizeCacheInvalid(void)
{
  sizeCacheInvalid = TRUE;
}


void wxMediaPasteboard::GetExtent(float *w, float *h)
{
  CheckRecalc();

  if (w)
    *w = totalWidth;
  if (h)
    *h = totalHeight;
}

Bool wxMediaPasteboard::ScrollTo(wxSnip *snip, float localx, float localy, 
				 float w, float h, Bool refresh, int bias)
{
  if (refresh)
    UpdateSnip(snip);
  return FALSE;
}

void wxMediaPasteboard::SetCaretOwner(wxSnip *snip)
{
  if (DoSetCaretOwner(snip))
    UpdateAll();
}


void wxMediaPasteboard::Resized(wxSnip *snip, Bool redraw_now)
{
  wxNode *node;
  wxSnipLocation *loc;

  if (!(node = snipLocationList->Find((long)snip)))
    return;

  if (!redraw_now)
    sequence++;
  BeginEditSequence();
  
  loc = (wxSnipLocation *)node->Data();  

  UpdateLocation(loc);

  loc->needResize = TRUE;
  needResize = TRUE;

  UpdateLocation(loc);

  EndEditSequence();
  if (!redraw_now)
    --sequence;
}


void wxMediaPasteboard::NeedsUpdate(wxSnip *snip, float localx, float localy, 
				    float w, float h)
{
  float x, y;

  GetSnipLocation(snip, &x, &y);
  Update(x + localx, y + localy, w, h);
}

Bool wxMediaPasteboard::ReleaseSnip(wxSnip *snip)
{
  return FALSE;
}

/************************************************************************/

float wxMediaPasteboard::ScrollLineLocation(long line, float displayHeight)
{
  return (displayHeight * line) / LINE_PP;
}


long wxMediaPasteboard::NumScrollLines(float displayHeight)
{
  float n;

  if (displayHeight <= totalHeight)
    return 0;

  n = (totalHeight * LINE_PP) / displayHeight;
  if (n != ((long) n))
    n += 1;

  return (long)n;
}

long wxMediaPasteboard::FindScrollLine(float y, float displayHeight)
{
  return (long)((y * LINE_PP) / displayHeight);
}

/************************************************************************/

void wxMediaPasteboard::SetMaxWidth(float w)
{
}

void wxMediaPasteboard::SetMinWidth(float w)
{
}

float wxMediaPasteboard::GetMaxWidth()
{
  return 0.0;
}

float wxMediaPasteboard::GetMinWidth()
{
  return 0.0;
}

void wxMediaPasteboard::SetMinHeight(float h)
{
  /* No nothing */
}

void wxMediaPasteboard::SetMaxHeight(float h)
{
  /* No nothing */
}

float wxMediaPasteboard::GetMinHeight()
{
  return 0.0;
}

float wxMediaPasteboard::GetMaxHeight()
{
  return 0.0;
}

/************************************************************************/


wxMediaBuffer *wxMediaPasteboard::CopySelf(void)
{
  wxMediaPasteboard *pb;

  pb = new wxMediaPasteboard();

  return pb;
}


float wxMediaPasteboard::GetDescent(void)
{
  return 0;
}


float wxMediaPasteboard::GetSpace(void)
{
  return 0;
}

void wxMediaPasteboard::GetCenter(float *fx, float *fy)
{
  float x, y, w, h;

  if (!admin) {
    w = totalWidth;
    h = totalHeight;
    x = y = 0;
  } else
    admin->GetView(&x, &y, &w, &h);

  if (fx)
    *fx = w / 2;
  if (fy)
    *fy = h / 2;

}

char *wxMediaPasteboard::GetFlattenedText()
{
  wxSnip *snip;
  char *t, *s, *old;
  long p, alloc, offset;

  alloc = 100;
  s = new char[alloc];

  snip = snips;

  p = 0;

  while (snip) {
    t = snip->GetText(0, snip->count, TRUE);

    offset = strlen(t);
    if (p + offset >= alloc) {
      alloc = 2 * (p + offset);
      old = s;
      s = new char[alloc];
      memcpy(s, old, p);
      delete[] old;
    }
    memcpy(s + p, t, offset);
    p += offset;
    delete[] t;

    snip = snip->next;
  }

  s[p] = 0;

  return s;
}

/************************************************************************/

void wxMediaPasteboard::Clear()
{
  Delete();
}

void wxMediaPasteboard::Cut(Bool extend, long time)
{
  Copy(extend, time);
  Clear();
}

void wxMediaPasteboard::Copy(Bool extend, long time)
{
  wxSnip *asnip;
  wxNode *node;
  wxSnipLocation *loc;

  if (!extend)
    FreeOldCopies(); 

  wxmb_commonCopyRegionData = NULL;

  for (node = snipLocationList->First(); node; node = node->Next()) {
    loc = (wxSnipLocation *)node->Data();
    if (loc->selected) {
      asnip = loc->snip->Copy();
      asnip->SetAdmin(NULL);
      wxmb_commonCopyBuffer->Append(asnip);
      wxmb_commonCopyBuffer2->Append(GetSnipData(loc->snip));
    }
  }

  InstallCopyBuffer(time);
}

void wxMediaPasteboard::Paste(long time)
{
  wxSnip *start, *snip;
  float cx, cy, left, right, top, bottom, dx, dy;
  wxSnipLocation *loc;
  wxDC *dc;

  start = snips;
  GetCenter(&cx, &cy);

  BeginEditSequence();

  NoSelected();

  DoPaste(time);

  if (admin && PTRNE(snips, start)) {
    dc = GetDC();

    /* Get top/left/bottom/right of pasted group: */
    for (snip = snips; PTRNE(snip, start); snip = snip->next) {
      loc = SnipLoc(snip);      
      if (loc->needResize)
	loc->Resize(dc);
      if (PTREQ(snip, snips)) {
	left = loc->x;
	top = loc->y;
	right = loc->r;
	bottom = loc->b;
      } else {
	if (loc->x < left)
	  left = loc->x;
	if (loc->y < top)
	  top = loc->y;
	if (loc->r > right)
	  right = loc->r;
	if (loc->b > bottom)
	  bottom = loc->b;
      }
      AddSelected(snip);
    }

    dx = cx - (left + right) / 2;
    dy = cy - (top + bottom) / 2;
    
    /* Shift the pasted group to center: */
    Move(dx, dy);
  } else {
    /* Just select them: */
    for (snip = snips; PTRNE(snip, start); snip = snip->next)
      AddSelected(snip);
  }

  EndEditSequence();
}

void wxMediaPasteboard::InsertPasteSnip(wxSnip *snip, wxBufferData *data)
{
  Insert(snip);
  SetSnipData(snip, data);
}

void wxMediaPasteboard::InsertPasteString(char *str)
{
  wxTextSnip *snip;

  snip = new wxTextSnip();
  snip->style = styleList->FindNamedStyle(STD_STYLE);
  snip->Insert(str, strlen(str));
  
  InsertPasteSnip(snip, NULL);
}

void wxMediaPasteboard::Kill(long time)
{
  Cut(time);
}

/************************************************************************/

Bool wxMediaPasteboard::GetSnipLocation(wxSnip *thesnip, float *x, float *y)
{
  wxNode *node;
  wxSnipLocation *loc;

  node = snipLocationList->Find((long)thesnip);
  if (!node)
    return FALSE;
  
  loc = (wxSnipLocation *)node->Data();
  if (x)
    *x = loc->x;
  if (y)
    *y = loc->y;

  return TRUE;
}

/************************************************************************/

wxBufferData *wxMediaPasteboard::GetSnipData(wxSnip *snip)
{
  wxNode *node;
  wxSnipLocation *loc;
  wxLocationBufferData *data;

  if (!(node = snipLocationList->Find((long)snip)))
    return wxMediaBuffer::GetSnipData(snip);

  loc = (wxSnipLocation *)node->Data();  
  data = new wxLocationBufferData;
  data->x = loc->x;
  data->y = loc->y;

  data->next = wxMediaBuffer::GetSnipData(snip);

  return data;
}

void wxMediaPasteboard::SetSnipData(wxSnip *snip, wxBufferData *data)
{
  while (data) {
    if (!strcmp(data->dataclass->classname, "wxloc")) {
      wxLocationBufferData *ldata;
      ldata = (wxLocationBufferData *)data;
      MoveTo(snip, ldata->x, ldata->y);
    }
    data = data->next;
  }
}

Bool wxMediaPasteboard::LoadFile(char *file)
{
  if (!file || !*file) {
    char *path;

    if (filename)
      path = copystring(PathOnly(filename));
    else
      path = NULL;

    file = wxFileSelector("Choose a file", path, NULL,
			  NULL, NULL, wxOPEN);

    if (path)
      delete[] path;
  }

  if (!file)
    return FALSE;

  if (::DirExists(file)) {
    wxMessageBox("Can't load a directory.", "Error");
    return FALSE;
  }

  SetFilename(file, FALSE);

  FILE *f = fopen(file, "rb");
  int n;
  
  if (!f) {
    wxMessageBox("Couldn't open the file.", "Error");
    return FALSE;
  }

  wxBeginBusyCursor();

  BeginEditSequence();

  Erase();

  char buffer[MRED_START_STR_LEN + 1];
  Bool fileerr;

  n = fread((char *)buffer, 1, MRED_START_STR_LEN, f);
  buffer[MRED_START_STR_LEN] = 0;
  if ((n != MRED_START_STR_LEN) || strcmp(buffer, MRED_START_STR)){
    wxMessageBox("This is not a MrEd file.", "Error");
    fileerr = TRUE;
  } else {
    fread((char *)wxme_current_read_format, 1, MRED_FORMAT_STR_LEN, f);
    fread((char *)wxme_current_read_version, 1, MRED_VERSION_STR_LEN, f);

    if (wxmeCheckFormatAndVersion()) {
      wxMediaStreamInFileBase b(f);
      wxMediaStreamIn mf(b);
      
      styleList->Clear();
      
      if (wxmbReadGlobalHeader(styleList, mf)) {
	if (mf.Ok())
	  fileerr = !ReadFromFile(mf);
	else
	  fileerr = TRUE;
      } else
	fileerr = TRUE;
      fileerr = !wxmbReadGlobalFooter(mf) || fileerr;
    
      styleList->NewNamedStyle(STD_STYLE, NULL);
      
      fileerr = fileerr || !mf.Ok();
    } else
      fileerr = TRUE;
  }

  fileerr = fileerr || ferror(f);
  
  fclose(f);

  if (fileerr)
    wxMessageBox("There was an error loading the file.", "Error");

  EndEditSequence();

  ClearUndos();

  wxEndBusyCursor();

  SetModified(fileerr);

  return !fileerr;
}

Bool wxMediaPasteboard::SaveFile(char *file)
{
  if (!file || !*file) {
    char *path, *pfile;

    if (filename) {
      path = copystring(PathOnly(filename));
      pfile = copystring(FileNameFromPath(filename));
    } else
      path = pfile = NULL;

    file = wxFileSelector("Save file as", path,
			  pfile, NULL, NULL, wxSAVE);

    if (path) {
      delete[] path;
      delete[] pfile;
    }
  }

  if (!file)
    return FALSE;

  FILE *f = fopen(file, "wb");
  Bool fileerr;

  if (!f) {
    wxMessageBox("Couldn't write the file.", "Error");
    return FALSE;
  }

  wxBeginBusyCursor();

  fwrite(MRED_START_STR, 1, MRED_START_STR_LEN, f);
  fwrite(MRED_FORMAT_STR, 1, MRED_FORMAT_STR_LEN, f);
  fwrite(MRED_VERSION_STR, 1, MRED_VERSION_STR_LEN, f);    

  wxMediaStreamOutFileBase b(f);
  wxMediaStreamOut mf(b);
  
  wxmbWriteGlobalHeader(styleList, mf);
  if (mf.Ok())
    fileerr = !WriteToFile(mf);
  wxmbWriteGlobalFooter(mf);
  
  fileerr = fileerr || !mf.Ok();

  fclose(f);

  if (fileerr)
    wxMessageBox("There was an error writing the file.", "Error");

  SetFilename(file, FALSE);

  wxEndBusyCursor();

  SetModified(fileerr);

  return !fileerr;
}

Bool wxMediaPasteboard::WriteToFile(wxMediaStreamOut &f)
{
  if (!DoWriteHeadersFooters(f, TRUE))
    return FALSE;

  wxmbWriteSnipsToFile(f, styleList, NULL, snips, NULL, NULL, this);
  
  if (!DoWriteHeadersFooters(f, FALSE))
    return FALSE;

  return TRUE;
}


Bool wxMediaPasteboard::ReadFromFile(wxMediaStreamIn &f)
{
  return ReadSnipsFromFile(f);
}


Bool wxMediaPasteboard::ReadInsert(wxSnip *snip)
{
  Insert(snip, (wxSnip *)NULL);
  return TRUE;
}

void wxMediaPasteboard::SetFilename(char *name, Bool temp)
{
  wxSnip *snip;

  if (filename)
    delete[] filename;
  filename = copystring(name);
  tempFilename = temp;

  for (snip = snips; snip; snip = snip->next)
    if (snip->flags & wxSNIP_USES_BUFFER_PATH)
      snip->SetAdmin(snipAdmin);
}

/************************************************************************/

void wxMediaPasteboard::StyleHasChanged(wxStyle *style)
{
  if (!style) {
    UpdateAll();
    return;
  }
}

/************************************************************************/

void wxMediaPasteboard::BeginEditSequence(Bool undoable)
{
  sequence++;
}


void wxMediaPasteboard::EndEditSequence(void)
{
  if (!(--sequence)) {
    sequenceStreak = FALSE;
    UpdateNeeded();
  }
}

/************************************************************************/

void wxMediaPasteboard::AddPasteboardFunctions(wxKeymap *tab)
{
}

Bool wxMediaPasteboard::GetDragable()
{
  return dragable;
}

void wxMediaPasteboard::SetDragable(Bool d)
{
  dragable = d;
}

/************************************************************************/

void wxMediaPasteboard::OnChange(void)
{
}

Bool wxMediaPasteboard::OnInsert(wxSnip *, wxSnip *, float x, float y)
{
  return TRUE;
}

void wxMediaPasteboard::AfterInsert(wxSnip *, wxSnip *, float x, float y)
{
}

Bool wxMediaPasteboard::OnDelete(wxSnip *)
{
  return TRUE;
}

void wxMediaPasteboard::AfterDelete(wxSnip *)
{
}

Bool wxMediaPasteboard::OnMoveTo(wxSnip *, float x, float y, Bool dragging)
{
  return TRUE;
}

void wxMediaPasteboard::AfterMoveTo(wxSnip *, float x, float y, Bool dragging)
{
}

Bool wxMediaPasteboard::OnResize(wxSnip *, float w, float h)
{
  return TRUE;
}

void wxMediaPasteboard::AfterResize(wxSnip *, float w, float h, Bool did)
{
}

/************************************************************************/

void wxMediaPasteboard::PrintToDC(wxDC *dc)
{
}

/************************************************************************/

void wxSnipLocation::Resize(wxDC *dc)
{
  snip->GetExtent(dc, x, y, &w, &h);
  r = x + w;
  b = y + h;
  hm = x + w/2;
  vm = y + h/2;
  
  needResize = FALSE;
}

